package LispList;
use strict;
use Symaxx2::Graphobject;
# LispList:
# ARGS contains a reference to another LispObject.

sub NewUnreal{
    my $self={};
    bless $self;    
    $::LispListCount++;# Memory-debug code
    my %P=@_;
    my $Command=$P{'OPERATOR'};
    my $String=$P{'STRING'};
    die unless (defined $Command || defined $String);
    # a * => a
    # a + => a
    if ($String){
	return &NewFromString('STRING' => $String,'UNREAL' => 1);
    };
    my @Data=@{$P{'DATA'}}; 
    if ($#Data==0 && ($Command eq 'MTIMES' || $Command eq 'MPLUS')){
	return $Data[0];
    };
    # ((MTIMES) $a $b ...)
    my $CommandLO=&NewFromString('STRING' => $Command,'UNREAL' => 1);
    $self->{'ARGS'}=[
		     &NewFromList('LIST' => [$CommandLO],'UNREAL' => 1),
		     @Data
		     ];
    $self->{'UNREAL'}=1;
    die unless $self->{'ARGS'}>=1;

    return $self;
};
sub NewFromList{
    my $self={};
    bless $self; 
    $::LispListCount++;# Memory-debug code
    my %P=@_;
    my $List=$P{'LIST'};die unless defined $List;
    my $Quoted=$P{'QUOTED'};
    my $Unreal=$P{'UNREAL'}; 
    $self->{'ARGS'}=$List;
    
    # debug code
    my @DebugList=@{$List};
    my $Elem;
    foreach $Elem(@DebugList){
	die $Elem unless ref($Elem) eq 'LispList';
    };
     #end debug code
  
    die unless $self->{'ARGS'}>=0;
    if ($Unreal){$self->{'UNREAL'}=1;};
    return $self;
};
sub NewFromString{
    my $self={};
    bless $self; 
    $::LispListCount++;# Memory-debug code
    my %P=@_;
    my $String=$P{'STRING'};die unless defined $String;
    my $Unreal=$P{'UNREAL'}; 
    my $Quoted=$P{'QUOTED'}; 
    $self->{'ARG1'}=$String;
    if ($Unreal){$self->{'UNREAL'}=1;};
    if ($Quoted){$self->{'QUOTED'}=1;};
    return $self;
};

sub QueryArgs{
    my $self=shift or die;
    return $self->{'ARGS'};
};
sub QueryCommand{
    my $self=shift or die;
    if (!defined $self->{'ARGS'}){return undef;};
    my $HeadElement=$self->{'ARGS'}[0]; die unless ref($HeadElement) eq 'LispList';
    my $Command=$HeadElement->QueryFirstListElement;
    return undef unless $Command;
    return $Command;
};
sub QueryFirstListElement{
    my $self=shift or die;
    if (!defined $self->{'ARGS'}){return undef;};
    my $HeadElement=$self->{'ARGS'}[0]; die unless ref($HeadElement) eq 'LispList';
    return $HeadElement->QueryAtomData;
};
sub QueryArrayFlag{
    my $self=shift or die;
    if (!defined $self->{'ARGS'}){return undef;};
    my @Args=@{$self->{'ARGS'}};
    my $Count; for ($Count=1;$Count<=$#Args;$Count++){
	my $Arg=$Args[$Count]->QueryAtomData;
	if ($Arg && $Arg eq 'ARRAY'){
	    return 1;
	};
    };
    return 0;
};
sub QueryStringList{
    # if this is a list of strings, return all the elements, separated by space.
    my $self=shift or die;
    my $Args=$self->{'ARGS'};
    if (!defined $Args){return undef;};
    my @ReturnList=();
    my $Arg; foreach $Arg(@{$Args}){
	my $Listelement=$Arg->QueryAtomData;
	return undef unless defined $Listelement;
	push @ReturnList,$Listelement;
    };
    return join(" ",@ReturnList);
};
sub QueryInverted{
    # gives back the inverted argument (x^-3 => 3)
    # works only for MEXPT
    my $self=shift or die;
    return undef unless defined $self->{'ARGS'}; # for example ATOM
    my @Args=@{$self->{'ARGS'}};
    return undef unless $#Args>=2; # for example MMINUS
    my $Command=$self->{'ARGS'}->[0] or die;
    my $Arg1=$self->{'ARGS'}->[1] or die;
    my $Arg2=$self->{'ARGS'}->[2] or die;
    if ($self->QueryCommand eq 'MEXPT'){
	# 10^-5 is not converted to 1/10^5 - doesn't make it easier to read
	if ($Arg1->QueryGeneralInfo->{'FLOAT'}||$Arg1->QueryGeneralInfo->{'INT'}){return undef;}
	my $NegArg2=$Arg2->QueryNegated; 
	return undef unless $NegArg2;
	my $NegArgData=$NegArg2->QueryAtomData;
	if ($NegArgData && $NegArgData eq '1'){ # x^-1
	    return $Arg1;
	} else {
	    my $Inverted=
		# Every result of algebraic manipulation done not in Maxima is UNREAL
	      LispList::NewFromList('LIST' => [$Command,$Arg1,$NegArg2],'UNREAL' => 1);
	    return $Inverted;
	};
    } # if MEXPT 
    elsif ($self->QueryCommand eq 'RAT'){
	my $Arg1String=$self->{'ARGS'}->[1]->QueryAtomData;
	if ($Arg1String && $Arg1String eq '1'){ # 1/x
	    return $self->{'ARGS'}->[2];          # return #
	};
    };  # if RAT
    return undef;
};
sub QueryNegated{
    my $self=shift or die;
    
    # special case: negative integer Atom
    if ($self->{'ARG1'}){
	if($self->{'ARG1'}=~/^\-(\d+)$/){
	    my $NegatedAtomNumber=$1;
	    return NewFromString('STRING' => $NegatedAtomNumber,'UNREAL' => 1);
	};
    };
    return undef unless defined $self->{'ARGS'}; # for example ATOM
    my @Args=@{$self->{'ARGS'}} or die;
    my $Arg0=$Args[0] or die;
    my $Command=$self->QueryCommand;
    if ($Command && $#Args>=1){ # NEG has one arg.
	my $Arg1=$Args[1] or die;
	if ($Command eq 'MTIMES' && $#Args>=2 ){ # MTIMES has two args
	    my $NegatedArg1=$Arg1->QueryNegated;
	    if ($NegatedArg1){
		my @NegatedArgList=();
		my $Count;for ($Count=2;$Count<=$#Args;$Count++){
		    push @NegatedArgList,$Args[$Count];
		};
		# special case 1*x*y => x*y
		my $NegArgAtomData=$NegatedArg1->QueryAtomData;
		if (!defined $NegArgAtomData || $NegArgAtomData ne '1'){
		    # add to head of list
		    unshift @NegatedArgList,$NegatedArg1;
		};
		if ($#NegatedArgList ==0){
		    return $NegatedArgList[0];
		};
		unshift @NegatedArgList,$Arg0; # (MTIMES) recycled
		return LispList::NewFromList('LIST' => \@NegatedArgList,
					     'UNREAL' => 1);
	    }; # if negated arg
	} elsif ($Command eq 'MMINUS'){
	    if ($#Args !=1){die 'MMINUS should have only one arg?!';};
	    return $Args[1];
	} elsif ($Command eq 'RAT'){
	    if ($#Args !=2){die 'RAT should have two args?!';};
	    my $NegatedArg1=$Arg1->QueryNegated;
	    if ($NegatedArg1){
		my @NegatedList=($Arg0,$NegatedArg1,$Args[2]);
#		print $self->QueryDebugString."\n";
#		print LispList::NewFromList('LIST' => \@NegatedList,
#       				    'UNREAL' => 1)->QueryDebugString."\n\n";
		return LispList::NewFromList('LIST' => \@NegatedList,
					     'UNREAL' => 1);
	    }; # if negated arg
	};
    }; # if $Command
    return undef;
};
sub QueryAtomData{
    my $self=shift or die;
    return $self->{ARG1};
};
sub GetGO{
    my $self=shift or die;
    my $GO=GO::New('LISPOBJECT' => $self,'PARENTOPERATOR' => 'NONE','LISPLINK'=>$self,@_); # pass on style option
    return $GO;
};
sub QueryDebugString{
    my $self=shift or die;
    if ($self->{'ARG1'}){return $self->{'ARG1'};};
    my @List;
    my $Elem;foreach $Elem(@{$self->{'ARGS'}}){
	if ($Elem && ref($Elem) eq 'LispList'){
	    push @List,$Elem->QueryDebugString;
	} elsif ($Elem){
	    push @List,"***PANIC: $Elem***";
	} else {
	    push @List,"***PANIC: UNDEFINED ***";
	}
    };
    return "(".join(":",@List).")";
};
sub QueryLispCode{
    my $self=shift or die;
    my %P=@_;
    my $ReturnString='';
    if ($self->{'QUOTED'}||($P{'FORCEQUOTE'})){
	$ReturnString.="\'"; # quotation mark
    };
    if (defined $self->{'ARG1'}){
	return $ReturnString.$self->{'ARG1'};
    };
    my @List;
    my $Elem;foreach $Elem(@{$self->{'ARGS'}}){
	push @List,$Elem->QueryLispCode;
    };
    return $ReturnString."(".join(" ",@List).")";
};
sub QueryListOfReferenceIDs{
    my $self=shift or die;
    my %List;
    if (defined $self->{'ARG1'}){return [];};
    my @Args=@{$self->{'ARGS'}};
    if ($#Args==1){ 
	my $StringList=$Args[0]->QueryStringList;
	# is this expression itself a reference?
	if ($StringList && $StringList =~/\|?\$REF\|?\s+ARRAY/){
	    my $RefTo=$Args[1]->QueryAtomData;
	    if (defined $RefTo){
		return [$RefTo]; # this is what the reference refers to
	    } else {
		return [];
	    };
	};
    };
    my $Arg; foreach $Arg(@Args){
	# Find references of each argument...
	my $ArgRefs=$Arg->QueryListOfReferenceIDs;
	# Add them to %List...
	if ($ArgRefs){
	    my $OneRef;foreach $OneRef(@{$ArgRefs}){
		# REF[a+1] for example gives undef -> skip it!
		if ($OneRef){
		    $List{$OneRef}++;
		};
	    };
	};
    }; # foreach argument
    my @AllRefs=keys(%List);
    return \@AllRefs;
};
sub QueryVariablesFromMaximaFile{
    my $self=shift or die;
    my $Arg;
    my $List={};
    foreach $Arg(@{$self->{'ARGS'}}){
	$Arg->AddVariableToListFromMaximaFile($List);
    };
    return $List;
};
sub AddVariableToListFromMaximaFile{
    my $self=shift or die;
    my $List=shift or die;
    return unless defined $self->{'ARGS'};
    my $Arg1AtomString=$self->{'ARGS'}[0]->QueryAtomData;
    if ($Arg1AtomString && $Arg1AtomString eq 'DSKSETQ'){
	my $Varname=$self->{'ARGS'}[1]->QueryAtomData;
	my $Object=$self->{'ARGS'}[2];
	$List->{$Varname}=$Object;
    };
};
sub QueryUnreal{
    my $self=shift or die;
    if ($self->{'UNREAL'}){return 1;}
    return 0;
};
sub QueryGeneralInfo{
    ######################################################################################
    # Generate or retrieve general information (of which is known, that it has to be     #
    # gnerated sooner or later, so we decide to do it sooner and in one go)              #
    ######################################################################################
    my $self=shift or die;

    # Retrieve saved copy?
    if ($self->{'GENERALINFO'}){return $self->{'GENERALINFO'};};

    my %GeneralInfo=();
    my $AtomData=$self->{'ARG1'};

    if (defined $AtomData){
    ######################################################################################
    # This is an atom                                                                    #
    ######################################################################################
	if ($AtomData eq '&'){$AtomData='|&|'}; # empty string
	if ($AtomData=~/^\-?\d+$/){
	    $GeneralInfo{'INT'}=1; 
	} elsif ($AtomData=~/^-?\.?\d/){ 
	    # Everything that starts with 
	    # -.1
	    # .1
	    # -1
	    # 1
	    # is considered a number (because Maxima symbols are not allowed to
	    # start with any of those characters).	
	    if ($AtomData=~/^\-?\d*\.\d*$/){
		$GeneralInfo{'FLOAT'}=1;  
		$AtomData=&FormatFloat($AtomData);
	    } elsif ($AtomData=~/^(\-*\d*\.\d*)[eE]\+?(-?\d+$)/){ 
		my $Mant=$1;my $Expt=$2;
		$Mant=&FormatFloat($Mant);
		$AtomData=$Mant.'E'.$Expt;
		$GeneralInfo{'FLOAT'}=1;  
	    } else {
		die "$AtomData is no recognized number format!";
	    }
	} else {
	    my $QuoteMe=0;
	    if ($AtomData=~/^\|(.*)\|$/){
		# |$Bla| to $Bla
		$AtomData=$1;
		$GeneralInfo{'PIPEDELIMITERS'}=1;	    
		$QuoteMe++;
	    };
	    if ($AtomData=~/^\$(.+)$/){
		# $Bla to Bla
		$AtomData='%'.$1; 
		$GeneralInfo{USEREXP}=1;	# Userexp is misleading: Could also be constant    
	    };
	    if ($AtomData=~/^\%(.+)$/){
		# %Bla to Bla
		$AtomData=$1;
		$GeneralInfo{'SYSTEMEXP'}=1;	    
	    };
	    if ($AtomData=~/^\&(.*)$/){
		# &Bla to Bla
		$AtomData=$1;
		$GeneralInfo{'AMPERSAND'}=1;	   
		$QuoteMe++;
	    };
	    if ($QuoteMe==2){ #|&xxxx|
		$AtomData='"'.$AtomData.'"';
	    };
	    if (!$GeneralInfo{'USEREXP'} && !$GeneralInfo{'SYSTEMEXP'} && !$GeneralInfo{'AMPERSAND'}){
		if ($AtomData eq 'T'){$AtomData='TRUE';};
		if ($AtomData eq 'NIL'){$AtomData='FALSE';};
	    };
	}; # if not float
	$GeneralInfo{'ATOMDATA'}=$AtomData;
    } else { 
	# if not atom
	die unless $self->{'ARGS'};
	# operator, function, array access or lisp list (MTIMES SIMP)
	my @Args=@{$self->{'ARGS'}};
	$GeneralInfo{'NRARGS'}=$#Args-1;
	die "Oops - less than two elements? " unless $#Args>=0;
	# $HeadElem is maybe (MTIMES SIMP)
	my $HeadElement=$Args[0] or die "Oops - fixme!";
	die unless ref($HeadElement) eq 'LispList';

	# Assume the first element is a list. The command has to be. 
	# (MTIMES a b) is not a valid statement.
	# ((MTIMES) a b) is.
	die 'Syntax ((operator ...) arg) violated! '.$self->QueryDebugString unless (defined $HeadElement->{'ARGS'});
	my $FirstArg=$HeadElement->{'ARGS'}[0];

	# Here we create the 'GeneralInfo' for the atom.
	my $FirstArgInfo=$FirstArg->QueryGeneralInfo;
	my $Cmd=$FirstArgInfo->{'ATOMDATA'};
	if ($Cmd){
	    $GeneralInfo{'COMMAND'}=$Cmd;
	    if ($FirstArgInfo->{'USEREXP'}){
		$GeneralInfo{'COMMANDUSEREXP'}=1;
		# in case of a userexpression ($Bla), we look for the ARRAY tag
		if ($HeadElement->QueryArrayFlag){
		    $GeneralInfo{'COMMANDARRAY'}=1;
		};
	    };
	    # Can't think of a reason, why somebody would want to access a system array.
	    if ($FirstArgInfo->{'SYSTEMEXP'}){$GeneralInfo{'COMMANDSYSTEMEXP'}=1;};
	} else {
  	    # if not Cmd
	    $GeneralInfo{'COMMAND'}='LISPLIST';
	}; # if not command
    }; # if not atom
    if ($self->{'UNREAL'}){$GeneralInfo{'UNREAL'}=1;};
    if ($self->{'QUOTED'}){$GeneralInfo{'QUOTED'}=1;};
    if (!$::ConserveMemory1){
	$self->{'GENERALINFO'}=\%GeneralInfo;
    };
    return \%GeneralInfo;
};
sub ReplaceObject{
    my $self=shift or die; 
    my %P=@_;
    my $Old=$P{'OLD'} or die;
    my $New=$P{'NEW'} or die;
    if (!$self->{'ARGS'}){return 0;}
    my @Args=@{$self->{'ARGS'}};
    my $Count; for ($Count=0;$Count<=$#Args; $Count++){
	if ($Args[$Count] eq $Old){
	    $self->{'ARGS'}->[$Count]=$New; return 1;
	} else {
	    if ($Args[$Count]->ReplaceObject('OLD' => $Old,'NEW' => $New)){
		return 1;
	    }; # if subobject found it
	};
    };   
    return 0;
};
sub PurgeRedundantData{
    my $self=shift or die; 

    # 'GENERALINFO' will be recreated the next time it's asked for.
    # This cuts down the filesize by almost 50 %!
    delete $self->{'GENERALINFO'};

    if (!$self->{'ARGS'}){return;}
    my @Args=@{$self->{'ARGS'}};
    my $Count; for ($Count=0;$Count<=$#Args; $Count++){
	$Args[$Count]->PurgeRedundantData;
    };
};
# test case: BLOCK(a:SCONCAT("echo \"set data style lines\; set title '",f,"' \; set term postscript eps\; set file ' ",EPSFile,"' \; plot '/tmp/gnu' \;pause 10 \" | gnuplot"))
sub NewFromLispString{
    my $self={};
    bless $self; 
    $::LispListCount++;# Memory-debug code
    my $LispString=shift;
    my $Offset=0;
    my $Length=length($LispString);
    my $FirstChar=' ';
    while ($FirstChar eq ' '){
	die unless $Offset<$Length;
	$FirstChar=substr($LispString,$Offset++,1);
    };
    if ($FirstChar eq "'"){$self->{'QUOTED'}=1;
			   die unless $Offset<$Length;
			   $FirstChar=substr($LispString,$Offset++,1);
		       }
    if ($FirstChar eq '('){
	my @DelimiterQueue=('(');
	my @StringmodeQueue=(0);
	# first non-blank character starts the contents.
	my $NextChar=' ';while ($NextChar eq ' '){
		die unless $Offset<$Length;
		$NextChar=substr($LispString,$Offset++,1);
	    };
	if ($NextChar eq "\\"){
	    die unless $Offset<$Length;
	    $NextChar=substr($LispString,$Offset++,1);
	};
	my $ArgString='';my $BracketsCount=1;my @Args=();
	my $Stringmode=0;
	
      CharLoop:while (1){
	  if ($NextChar eq '|'){
	      if ($Stringmode eq '|'){
		  $Stringmode=0;
	      }elsif (!$Stringmode){
		  $Stringmode='|';
	      }; 
	  } elsif ($NextChar eq '"'){
	      if ($Stringmode eq '"'){
		  $Stringmode=0;
	      }elsif (!$Stringmode){
		  $Stringmode='"';
	      }; 
	  } elsif (!$Stringmode && $NextChar eq '('){
	      $BracketsCount++;
	  } elsif (!$Stringmode && $NextChar eq ')'){
	      $BracketsCount--;
	  } elsif ($NextChar eq ' ' && $ArgString eq ''){$NextChar='';} # don't append one trailing space
	  if (!$Stringmode && $NextChar eq ')' && $BracketsCount==0){last CharLoop;}
	  if (!$Stringmode && $NextChar eq ' ' && $BracketsCount==1){
	      push @Args,&NewFromLispString($ArgString);$ArgString='';
	  } else {
	      $ArgString.=$NextChar;
	  };
	  die unless $Offset<$Length;
	  $NextChar=substr($LispString,$Offset++,1);
	  if ($NextChar eq "\\"){
	      die unless $Offset<$Length;
	      $NextChar.=substr($LispString,$Offset++,1);
	  };
      }; # while next char	
	push @Args,&NewFromLispString($ArgString);
	
	$self->{'ARGS'}=\@Args;
	die unless $Offset == $Length;
    } elsif ($FirstChar eq '|' || $FirstChar eq '"'){
	my $Delimiter=$FirstChar;
#	print "New $Delimiter---$Delimiter       X".join("",@AllChars)."X\n";
	$self->{'ARG1'}=$Delimiter.substr($LispString,$Offset,$Length-$Offset);
	die unless $self->{'ARG1'}=~/^$Delimiter.*$Delimiter$/;
    } else {
	$self->{'ARG1'}=$FirstChar.substr($LispString,$Offset,$Length-$Offset);
    };
#    print "Returning ".$self->QueryDebugString."\n";
    return $self;
};
sub FormatFloat{
    my $Nr=shift; die unless defined $Nr;
    if ($::DoNotRound){return $Nr;};
    # We tip over the last '9' here to turn .xxx99999 to .yyy000000.
    $Nr=sprintf("%.13f",$Nr); # Maxima has 16 digits accuracy. 
    $Nr=~s/0+$//; # chop off trailing zeroes created by rounding
    $Nr=~s/000000\d$//g; # chop off numerical artifacts 
    $Nr=~s/0+$//; # chop off trailing zeroes again
    return $Nr;
};
sub QueryINPARTPathToLispObject{
    my $self=shift or die;
    my %P=@_;
    my $Lispobject=$P{'LISPOBJECT'};
    # We have to catch UNDEF here, because otherwise (almost) any object agrees with UNDEF.
    if (!$Lispobject){return undef;}
    if ($self eq $Lispobject){return [];}
    if (!defined $self->{'ARGS'}){return undef;};
    my @Args=@{$self->{'ARGS'}};
    my $Arg; my $ArgCount;
    for ($ArgCount=0;$ArgCount<=$#Args;$ArgCount++){
	$Arg=$Args[$ArgCount] or die;
	my $ArgPath=$Arg->QueryINPARTPathToLispObject('LISPOBJECT' => $Lispobject);
	if (defined $ArgPath){ # will be [] for a direct hit
	    unshift @{$ArgPath},$ArgCount;
	    return $ArgPath;
	};
    };
    return undef;    
};

sub DESTROY{
    # Memory-debug code
    $::LispListCount--;
}
sub QueryObjectToPath{
    my $self=shift or die;
    my @Path=@_;
    if ($#Path ==-1){return $self;} 
    if (!defined $self->{'ARGS'}){
#die $self->{'ARG1'}." ".join(",",@Path); 
	return undef;} # this die has to go
    my @Args=@{$self->{'ARGS'}};
    my $ArgNum=shift @Path;

    # Maxima-INPART index 1 corresponds to perl-index of 1, 
    # because Args-element 0 contains the header.
    if ($ArgNum>$#Args){return undef;}
    return $Args[$ArgNum]->QueryObjectToPath(@Path);
};
1;

