package MainDialog;
use Tk;
use strict;
use Symaxx2::PopupWindow;
@MainDialog::ISA=('PopupWindow');
# General form of widget hierarchy:
# The data array is an array of data lines
# One data line: [Keyword [DATA] {Args}]
# Data can be left empty. Args could be omitted.
sub New{
    my $proto = shift;
    my %P=@_;
    my $class = ref($proto) || $proto;

    # The flags are collected into three lists, sorted according to related
    # data structure. If something higher is selected, the lower flags
    # are not updated / set.
    my $ThingListOfFlags=['SHOWHEAD','SHOWDOC','SHOWCONTENTS','INTERFACE','DEFINITION','RECALLNOEVAL','EVALATLOAD'];
    my $LineListOfFlags=['SHOWCOMMAND','SHOWRESULT','SKIPEVAL','DONOTSIMP','NUMER','EVAL','NODISPFORM','SEL_SELSIMP','GNUP_CREATEDATA','GNUP_INDEP','GNUP_MIN','GNUP_MAX','GNUP_PTS','GNUP_SMOOTH','GNUP_CURVETITLE','GNUP_PLOTDATA','GNUP_CURVELOG','GNUP_DIAGTITLE','GNUP_DIAGLOGX','GNUP_DIAGLOGY','GNUP_DIAGGRIDX','GNUP_DIAGGRIDY','GNUP_DIAGSCALE','GNUP_DIAGLABELX','GNUP_DIAGLABELY','GNUP_DIAGRANGEX','GNUP_DIAGRANGEY','GNUP_DIAGCOLOUR','GNUP_LISTPLOT'];
    my $CommandListOfFlags=['SEL_SELECT','SEL_UNSELECT','SEL_PICK','UNIT_MULTIPLY','UNIT_CONVERT'];

    # WINPROP: Set window properties (title)
    my $WinProps=['WINPROP',[],{'TITLE' => 'Properties'}],
    
    my %F=();
    my $OuterNotebook=&Generate_Outer_Notebook(FLAGS => \%F);
    # Final assembly of the data structure
    my $Data=[$WinProps,$OuterNotebook]; # Data
    
    # Now we create a Popup window object using the data structure
    my $self=$class->SUPER::New
	(DATA => $Data,
	 FLAGS_THING => $ThingListOfFlags,
	 FLAGS_LINE => $LineListOfFlags,
	 FLAGS_COMMAND => $CommandListOfFlags,

         # It is more practical, that clicking on a part of the line or clicking on the line
         # produces the LINE page. Therefore, Line appears twice.
	 RAISE_THING => 'Whole object', # name of page
	 RAISE_LINE => 'Line', # name of page
	 RAISE_COMMAND => 'Line', # name of page
	 %P);

    $self->{FLAGS}=\%F;
    return $self;
}; # sub New
sub Generate_Outer_Notebook{
    my %P=@_; my $F=$P{FLAGS} or die;
    # There are some special lines: 
    # SETSELTYPE tells the window, that all following widgets are related
    # to a selected Thing / Line / Command.
    # If something higher is selected, the widgets will be disabled.
    my $SelTypeToThing=['SET_SELTYPE',[],{'SELTYPE' => 'THING'}];
    my $SelTypeToLine=['SET_SELTYPE',[],{'SELTYPE' => 'LINE'}];
    my $SelTypeToCommand=['SET_SELTYPE',[],{'SELTYPE' => 'COMMAND'}];
    
    my $ThingNotebookPage=&Generate_Thing_NotebookPage(FLAGS => $F);
    my $LineNotebookPage=&Generate_Line_NotebookPage(FLAGS => $F);
    my $CommandNotebookPage=&Generate_Command_NotebookPage(FLAGS => $F);
    my $OuterNotebook=['NOTEBOOK',
		       [$SelTypeToThing,      # instructions
			$ThingNotebookPage,   # data
			$SelTypeToLine,       # instructions
			$LineNotebookPage,    # data
			$SelTypeToCommand,    # instructions
			$CommandNotebookPage  # data
			], {} ];
    return $OuterNotebook;
};
sub Generate_Thing_NotebookPage{
    my %P=@_; my $F=$P{FLAGS} or die;
    my $Thing_DisplayPage=&Generate_Thing_DisplayPage(FLAGS => $F) or die;
    my $Thing_EvalPage=&Generate_Thing_Eval_Page(FLAGS => $F) or die;
    
    my $ThingNotebook=
	['NOTEBOOK',[$Thing_DisplayPage,$Thing_EvalPage], 
	 {-balloon => "Properties of the whole object\nDisplay:Turn on /off display of parts
Eval: Affects the way Maxima evaluates the command\n"} ];
    my $ThingNotebookPage=['PAGE',[$ThingNotebook],{-label => 'Whole object'}];
    return $ThingNotebookPage;
};

sub Generate_Line_NotebookPage{
    my %P=@_; my $F=$P{FLAGS};

    my @LinePagelist=();
    push @LinePagelist, &Generate_Line_Display_Page(FLAGS => $F);
    push @LinePagelist, &Generate_Line_Eval_Page(FLAGS => $F);
    push @LinePagelist, &Generate_Line_Refs_Page(FLAGS => $F);

    if ($::UIManager->QueryFeature(FEATURE => 'SELECTION')){
	push @LinePagelist,&Generate_Line_SelectionPage(FLAGS => $F);
    };
    if ($::UIManager->QueryFeature(FEATURE => 'GNUPLOT')){  
	push @LinePagelist,&Generate_Line_GnuplotDataPage(FLAGS => $F);
	push @LinePagelist,&Generate_Line_GnuplotPlotPage(FLAGS => $F);
    };

    my $LineNotebook=['NOTEBOOK',\@LinePagelist, {-balloon => "Properties of one command / result line
Display: Show/don't show command/result...
Eval:Flags that affect the evaluation of this line
Refs:References to include in the ev()-command for this line
GnuplotCurve:Create a data object for Gnuplot (one curve)
GnuplotDiag:Run Gnuplot and create an image"} ];
    my $LineNotebookPage=['PAGE',[$LineNotebook],{-label => 'Line'}];
    return $LineNotebookPage;
};

sub Generate_Command_NotebookPage{
    my %P=@_; my $F=$P{FLAGS} or die;
    my @CommandPagelist=();
    if ($::UIManager->QueryFeature(FEATURE => 'SELECTION')){
	push @CommandPagelist,&Generate_Command_SelectionPage(FLAGS => $F); 
    };
    
    if ($::UIManager->QueryFeature(FEATURE => 'UNITS')){
	push @CommandPagelist,&Generate_Command_UnitsPage(FLAGS => $F);
    };

    my $CommandNotebook=['NOTEBOOK',\@CommandPagelist, {-balloon => "Modifications on the current command or subexpression"} ];
    my $CommandNotebookPage=['PAGE',[$CommandNotebook],{-label => 'Command'}];
    return $CommandNotebookPage;
};

sub Generate_Thing_DisplayPage{
    my %P=@_; my $F=$P{FLAGS} or die;
    my $DisplayHead=['CB',[],{-text => 'Show header',-variable => \$F->{'SHOWHEAD'},-balloon => 'Toggles, if the header (containing the unique identifier number and some flags) is shown.'}];
    my $DisplayContents=['CB',[],{-text => 'Show contents',-variable => \$F->{'SHOWCONTENTS'},-balloon => 'Toggles, whether the contents (commands / results) are shown'}];
    my $DisplayDoc=['CB',[],{-text => 'Show documentation',-variable => \$F->{'SHOWDOC'},-balloon => 'Toggles, if the documentation entered into the doc-field is shown'}];
    my $DisplayInterface=['CB',[],{-text => 'Show also on interface page',-variable => \$F->{'INTERFACE'},-balloon => 'Toggles, if this "thing" appears on the 
"interface page". You can move it there to another 
location. Use this to group input and output of your
 calculation without all the mess in between.'}];
    my $DisplayPage=['PAGE',[$DisplayHead,$DisplayContents,$DisplayDoc,$DisplayInterface],{-label => 'Display'}]; #page
    return $DisplayPage;
}; # Generate_Thing_Display_Page



sub Generate_Thing_Eval_Page{
    my %P=@_; my $F=$P{FLAGS} or die;
    my $EvalDefinition=['CB',[],{-text => 'Definition',-variable => \$F->{'DEFINITION'},-balloon => "Toggles, if this 'thing' must be present,
before anything else is evaluated.
This can be used for
- a function definition (myfunc(x):=x^2 or myfunc(x)::=x^2),
- a variable definition (MyVar:=42),
- assumptions (assume(x>0)),
- for setting global Maxima switches (DEMOIVRE:TRUE) etc... ."}];
    my $EvalDoNotEval=['CB',[],{-text => 'Do NOT evaluate',-variable => \$F->{'RECALLNOEVAL'}}];
    my $EvalAtLoad=['CB',[],{-text => 'Eval automatically when loading file',-variable => \$F->{'EVALATLOAD'},-balloon => 'If this switch is set, the command will be evaluated
each time the file is loaded. The demo files use this to 
create plots.'}];
    my $EvalPage=['PAGE',[$EvalDefinition,$EvalDoNotEval,$EvalAtLoad],{-label => 'Eval'}]; #page
    return $EvalPage;
}; # Generate_Thing_Eval_Page
   
sub Generate_Line_Display_Page{    
    my %P=@_; my $F=$P{FLAGS} or die;
    my $L_DisplayCommand=['CB',[],{-text => 'Show command',-variable => \$F->{'SHOWCOMMAND'},-balloon => 'Toggles, whether the command (which is fed into Maxima) is shown'}];
    my $L_DisplayResult=['CB',[],{-text => 'Show result',-variable => \$F->{'SHOWRESULT'},-balloon => 'Toggles, whether the result (which is returned by Maxima) is shown'}];
    my $L_DisplayPage=['PAGE',[$L_DisplayCommand,$L_DisplayResult],{-label => 'Display'}];
    return $L_DisplayPage;
};
sub Generate_Line_Eval_Page{
      my %P=@_; my $F=$P{FLAGS} or die;
  my $L_EvalSkipEval=['CB',[],{-text => 'Skip evaluation',-variable => \$F->{'SKIPEVAL'},-balloon => "In most cases: OFF!
The command is evaluated when you hit the eval button. 
For example, 1+1 gives 2.
Set this switch and the evaluation is NOT performed.
1+1 will remain 1+1 in this case. REF[]-statements will
still be evaluated, though.
Useful for manually expanding expressions.
"}];
    my $L_EvalDoNotSimp=['CB',[],{-text => 'Do NOT simplify',-variable => \$F->{'DONOTSIMP'},-balloon => "In most cases: OFF!
When an expression is evaluated, it has to be simplified.
This is Maxima's default behaviour when evaluating.
You can turn this off here. Same as
block([simp:off],MyCommand)"}];
    my $L_EvalNumer=['CB',[],{-text => 'Eval to number',-variable => \$F->{'NUMER'},-balloon => 'Toggles, if the command is evaluated using ev(cmd) (disabled)
or ev(cmd,NUMER) (enabled).
This will result in floating-point output, if possible.
Otherwise, fractions will be returned.'}];  
    my $L_EvalEval=['CB',[],{-text => 'Repeated eval',-variable => \$F->{'EVAL'},-balloon => 'Toggles, if the command is evaluated using ev(cmd) (disabled)
or ev(cmd,EVAL) (enabled).
In general, the EVAL switch repeats the evaluation, until
the result does not change anymore.'}];  
    my $L_EvalNoDispForm=['CB',[],{-text => "Don't apply dispform()",-variable => \$F->{'NODISPFORM'},-balloon => "Usually: OFF!
By default, Symaxx will convert output from Maxima 
into a form, that is convenient to read. 
However, there are cases where Maxima's
internal representation is more useful (when working 
with INPART). This switch will disable the DISPFORM 
processing of output."}]; 
   my $L_EvalPage=['PAGE',[$L_EvalSkipEval,$L_EvalDoNotSimp,$L_EvalNumer,$L_EvalEval,$L_EvalNoDispForm],{-label => 'Eval'}];  
    return $L_EvalPage;
}; # Generate_Line_Eval_Page

sub Generate_Line_Refs_Page{
    my %P=@_; my $F=$P{FLAGS} or die;

    # Code
    my $RefsListboxSetContext=sub{
	my %P=@_;
	my $Widget=$P{WIDGET} or die;
	my $C=$P{CONTEXT} or die;
	my %Context=%{$C};
	$Widget->delete(0,'end');
	my $Thing=$Context{'THING'}; return unless $Thing;
	my $Index=$Context{'INDEX'}; return unless defined $Index;
	my $ID=$Thing->QueryID;
	my $RefsInEval=$Thing->QueryFlag
	    (FLAGMODE => 'LINE',INDEX => $Index,FLAG => 'REFSINEVAL');  
	if ($RefsInEval){
	    my $ID;foreach $ID(@{$RefsInEval}){
		my $ReferencedThing=$::ThingManager->QueryThing('ID' => $ID,'TOLERATEINVALID'=>1);
		my $NewListboxLine;
		if ($ReferencedThing){
		    my $DocLine=$ReferencedThing->QueryDoc;
		    $NewListboxLine=$ID.': '.$DocLine;
		} else {
		    $NewListboxLine=$ID.': ###does not exist###';
		    };
		$Widget->insert('end',$NewListboxLine);
	    };
	}; # if refs in eval
    };
    my $RefsEntryLineSetContext=sub{	
	my %P=@_;
	my $Widget=$P{WIDGET} or die;
	my $C=$P{CONTEXT} or die;
	my %Context=%{$C};
	
	$F->{REFSINEVAL_TEXT}='';
	
	my $Thing=$Context{'THING'}; return unless $Thing;
	my $Index=$Context{'INDEX'}; return unless defined $Index;
	my $RefsInEval=$Thing->QueryFlag
	    (FLAGMODE => 'LINE',INDEX => $Index,FLAG => 'REFSINEVAL');  
	return unless $RefsInEval;
	$F->{REFSINEVAL_TEXT}=join(",",@{$RefsInEval});
    };
    my $RefsEntryLineUpdate=sub{
	my %P=@_;
	my $Widget=$P{WIDGET} or die;
	my $C=$P{CONTEXT} or die;
	my %Context=%{$C};

	my $Textline=$F->{REFSINEVAL_TEXT};
	my $Thing=$Context{THING}; return unless $Thing;
	my $Index=$Context{INDEX}; return unless defined $Index;
	# Special rule here: If you read REFSINEVAL,
	# it is an array. But if you set it, it's a string.
	$Thing->SetFlag
	    (FLAGMODE => 'LINE',INDEX => $Index,FLAG => 'REFSINEVAL','NEW' => $Textline);  
    };


    my $L_RefsListbox=['LB',[],{SETCODE => $RefsListboxSetContext,-balloon => 'This shows all the other expressions, which are given
as arguments to the ev()-command. Also their documentation 
is shown to help identify them. Change them in the line below.'}];
    my $L_RefsEntry=['E',[],{-text => 'Refs in eval',-variable => \$F->{'REFSINEVAL_TEXT'},SETCODE => $RefsEntryLineSetContext ,UPDATECODE => $RefsEntryLineUpdate,-balloon => 'A list of the expressions, which are
given as arguments to the ev()-command.
For example, if you put here
14,15,16
then the command will be evaluated using
ev(cmd,REF[14],REF[15],REF[16])
Using this line as a command and leaving
this field empty will give the same result,
only the command looks more confusing.
'}];

    # Selection for line

    my $L_RefsPage=['PAGE',[$L_RefsListbox,$L_RefsEntry],{-label => 'Refs'}];
    return $L_RefsPage;
};# Generate_Line_Refs_Page
    
    sub Generate_Line_SelectionPage{
	my %P=@_; my $F=$P{FLAGS} or die;
my $L_SelSimpCB=['CB',[],{-text => 'Apply SelSimp',-variable => \$F->{'SEL_SELSIMP'},-balloon => 'Applies the (Symaxx)-SelSimp command, which tries to group
different selections belonging to the same group together.'}]; 
	return ['PAGE',[$L_SelSimpCB],{-label => 'Sel'}];
    };




sub Generate_Line_GnuplotDataPage{
    my %P=@_; my $F=$P{FLAGS} or die;
    # Gnuplot for line

	
	my $EnableDisableAllPlotStuff=sub{
	    # This sub is called, when the state of the window changes.
	    # We enable / disable the widget here.
	    # Since this overrides the higher-level mechanism, we'll have to check for that
	    # again (Contexttype).
	    # The Entry consists of a label and an entry line.
	    # To make it clear to the user, that the line is disabled, the text
	    # is grayed out. That's what the LABEL is good for.
	    my %P=@_;
	    my $Widget=$P{WIDGET} or die;
	    my $C=$P{CONTEXT} or die;
	    my %Context=%{$C};
	    my $Label=$P{LABEL}; # only defined for entry & label
	    
	    my $NewState='normal'; my $Color='black';
	    if (!$F->{GNUP_CREATEDATA} || $Context{CONTEXTTYPE} ne 'COMMAND'){$NewState='disabled';$Color='darkgray';};
	    $Widget->configure(-state => $NewState);
	    if ($Label){
		$Label->configure(-foreground => $Color);
	    };
	};
	my $EnableDisableNonListPlotStuff=sub{
	    my %P=@_;
	    my $Widget=$P{WIDGET} or die;
	    my $C=$P{CONTEXT} or die;
	    my $Label=$P{LABEL}; # only defined for entry & label

	    my %Context=%{$C};
	    
	    my $NewState='normal';my $Color='black';
	    if (!$F->{GNUP_CREATEDATA} || $F->{GNUP_LISTPLOT} || $Context{CONTEXTTYPE} ne 'COMMAND'){$NewState='disabled';$Color='darkgray';};
	    $Widget->configure(-state => $NewState);
	    if ($Label){
		$Label->configure(-foreground => $Color);
	    };
	    
	};
	
	my $L_Gnuplot_Enable=['CB',[],{-text => 'Create Gnuplot data',-variable => \$F->{GNUP_CREATEDATA},-balloon => 'Create a new Gnuplot data object
(basically a list of points)
    This option is OFF by default.'}]; 
    my $L_Gnuplot_Indep=['E',[],{-text => 'Independent',-variable => \$F->{GNUP_INDEP},-balloon => "The variable on the horizontal axis",
 SETCODE => $EnableDisableNonListPlotStuff}];
    my $L_Gnuplot_Min=['E',[],{-text => 'Min',-variable => \$F->{GNUP_MIN},-balloon => "Start of plotting range",
 SETCODE => $EnableDisableNonListPlotStuff}];
    my $L_Gnuplot_Max=['E',[],{-text => 'Max',-variable => \$F->{GNUP_MAX},-balloon => "End of plotting range",
 SETCODE => $EnableDisableNonListPlotStuff}];
    my $L_Gnuplot_Pts=['E',[],{-text => 'Nr. points',-variable => \$F->{GNUP_PTS},-balloon => "How many points to plot",
 SETCODE => $EnableDisableNonListPlotStuff}];
    my $L_Gnuplot_Title=['E',[],{-text => 'Curve title',-variable => \$F->{GNUP_CURVETITLE},-balloon => "Title of this curve",
 SETCODE => $EnableDisableAllPlotStuff}];
    my $L_Gnuplot_Smooth_none=['RB',[],{-text => "Points",
			   -variable => \$F->{GNUP_SMOOTH},
			   -value => '',-balloon => 'Plot points only',
 SETCODE => $EnableDisableNonListPlotStuff}];
    my $L_Gnuplot_Smooth_unique=['RB',[],{-text => "Straight lines",
			   -variable => \$F->{GNUP_SMOOTH},
			   -value => 'unique',
-balloon => 'Connect data points with
straight lines',SETCODE => $EnableDisableNonListPlotStuff}];
    my $L_Gnuplot_Smooth_bezier=['RB',[],{-text => "Bezier Curves",
			   -variable => \$F->{GNUP_SMOOTH},
			   -value => 'bezier',
-balloon => 'Connect data points with
bezier curves',SETCODE => $EnableDisableNonListPlotStuff}];
    my $L_Gnuplot_Smooth_cspline=['RB',[],{-text => "CSpline",
			   -variable => \$F->{GNUP_SMOOTH},
			   -value => 'cspline',
-balloon => 'Connect data points with
cspline curves',SETCODE => $EnableDisableNonListPlotStuff}];
    my $L_Gnuplot_Smooth_sbezier=['RB',[],{-text => "SBezier",
			   -variable => \$F->{GNUP_SMOOTH},
			   -value => 'sbezier',
-balloon => 'Connect data points with
SBezier curves',SETCODE => $EnableDisableNonListPlotStuff}];




#    my $L_Gnuplot_Smooth=['OM',[],{
#			   -variable => \$F->{GNUP_SMOOTH},
#			   -options => [['Points only',''],['straight lines','unique'],['Bezier curves','bezier'],['CSplines','cspline'],['SBezier curves','sbezier']],
# -balloon => 'Interpolate between data points?'}];
 my $L_Gnuplot_Log=['CB',[],{-text => 'x-axis log',-variable => \$F->{GNUP_CURVELOG},-balloon => "Increment the independent variable
on a logarithmic scale.
Don't forget to turn on logarithmic scale
when plotting, too!",SETCODE => $EnableDisableNonListPlotStuff}]; 
    my $L_Gnuplot_ListPlot=['CB',[],{-text => 'Plot existing list',-variable => \$F->{GNUP_LISTPLOT},
-balloon => "The data is an existing list [[x0,y0],[y1,y1],...].
    The independent variable is not used in this case.",SETCODE => $EnableDisableAllPlotStuff}];
	my $L_GnuplotDataPage=['PAGE',[$L_Gnuplot_Enable,
$L_Gnuplot_ListPlot,$L_Gnuplot_Indep,$L_Gnuplot_Min,$L_Gnuplot_Max,$L_Gnuplot_Pts,$L_Gnuplot_Log,
$L_Gnuplot_Smooth_none,
$L_Gnuplot_Smooth_unique,
$L_Gnuplot_Smooth_bezier,
$L_Gnuplot_Smooth_cspline,
$L_Gnuplot_Smooth_sbezier,
$L_Gnuplot_Title],{-label => 'Gnuplot curve'}]; 
return $L_GnuplotDataPage;
}; # Generate_Line_GnuplotDataPage

sub Generate_Line_GnuplotPlotPage{
    my %P=@_; my $F=$P{FLAGS} or die;
    my $CreateGnuPlotStuff=sub{
	my %P=@_;
	my $Widget=$P{WIDGET} or die;
	my $C=$P{CONTEXT} or die;
	my $Label=$P{LABEL}; # only defined for entry & label
	
	my %Context=%{$C};
	
	my $NewState='normal';my $Color='black';
	if (!$F->{GNUP_PLOTDATA} || $Context{CONTEXTTYPE} ne 'COMMAND'){$NewState='disabled';$Color='darkgray';};
	$Widget->configure(-state => $NewState);
	if ($Label){
	    $Label->configure(-foreground => $Color);
	};
    };
    
  	my $L_Gnuplot_PlotData=['CB',[],{-text => 'Create plot from data',-variable => \$F->{GNUP_PLOTDATA},-balloon => 'Run data object(s) through GNUPLOT'}];  
my $L_Gnuplot_LogX=['CB',[],{-text => 'x-axis log',-variable => \$F->{GNUP_DIAGLOGX},-balloon => "X-axis logarithmic scale",SETCODE => $CreateGnuPlotStuff}]; 
my $L_Gnuplot_LogY=['CB',[],{-text => 'y-axis log',-variable => \$F->{GNUP_DIAGLOGY},-balloon => "Y-axis logarithmic scale",SETCODE => $CreateGnuPlotStuff}]; 
my $L_Gnuplot_GridX=['CB',[],{-text => 'x Grid',-variable => \$F->{GNUP_DIAGGRIDX},-balloon => "X-axis grid",SETCODE => $CreateGnuPlotStuff}]; 
my $L_Gnuplot_GridY=['CB',[],{-text => 'y Grid',-variable => \$F->{GNUP_DIAGGRIDY},-balloon => "Y-axis grid",SETCODE => $CreateGnuPlotStuff}]; 
my $L_Gnuplot_Colour=['CB',[],{-text => 'Colour',-variable => \$F->{GNUP_DIAGCOLOUR},-balloon => "B/W or colour",SETCODE => $CreateGnuPlotStuff}]; 
my $L_Gnuplot_RangeX=['E',[],{-text => 'x-Range',-variable => \$F->{GNUP_DIAGRANGEX},-balloon => "X-axis range.
1:3 - set lower and upper limit to 1 and 3
:3 - set upper limit to 3
1: - set lower limit to 1
Note: The plot MUST contain at least
one data point!",SETCODE => $CreateGnuPlotStuff}]; 
my $L_Gnuplot_RangeY=['E',[],{-text => 'y-Range',-variable => \$F->{GNUP_DIAGRANGEY},-balloon => "Y-axis range.
1:3 - set lower and upper limit
:3 - set upper limit
3: - set lower limit
Note: The plot MUST contain at least
one data point!",SETCODE => $CreateGnuPlotStuff}]; 

    my $L_Gnuplot_DiagTitle=['E',[],{-text => 'Diagram title',-variable => \$F->{GNUP_DIAGTITLE},-balloon => "Title of this diagram",SETCODE => $CreateGnuPlotStuff}];
    my $L_Gnuplot_DiagScale=['E',[],{-text => 'Scale',-variable => \$F->{GNUP_DIAGSCALE},-balloon => "Scale (1 => 100 %)",SETCODE => $CreateGnuPlotStuff}];
    my $L_Gnuplot_DiagLabelX=['E',[],{-text => 'X-label',-variable => \$F->{GNUP_DIAGLABELX},-balloon => "x-axis label",SETCODE => $CreateGnuPlotStuff}];
    my $L_Gnuplot_DiagLabelY=['E',[],{-text => 'Y-label',-variable => \$F->{GNUP_DIAGLABELY},-balloon => "y-axis label",SETCODE => $CreateGnuPlotStuff}];

     my $L_GnuplotPlotPage=['PAGE',[$L_Gnuplot_PlotData,$L_Gnuplot_DiagTitle,$L_Gnuplot_DiagScale,$L_Gnuplot_DiagLabelX,$L_Gnuplot_DiagLabelY,
$L_Gnuplot_LogX,$L_Gnuplot_LogY,$L_Gnuplot_GridX,$L_Gnuplot_GridY,$L_Gnuplot_Colour,$L_Gnuplot_RangeX,$L_Gnuplot_RangeY],{-label => 'Gnuplot diagram'}]; 
 	return $L_GnuplotPlotPage,
	
    }; # Generate_Line_GnuplotPlotPage



sub Generate_Command_SelectionPage{
    my %P=@_; my $F=$P{FLAGS} or die;

    my $C_SelRB1=['RB',[],{-text => "Don't do anything (default)",
			   -variable => \$F->{'APPLY_SEL_MODE'},
			   -value => '',-balloon => 'Default.
The "Selection" feature is not used on this (sub)-expression.'}];
    my $C_SelRB2=['RB',[],{-text => 'Add Selection',
			   -variable => \$F->{'APPLY_SEL_MODE'},
			   -value => 'SELECT',-balloon => 
"Apply selection on this (sub)-expression
This will wrap it into a SEL()-function.
Those are taken out of the data each time
an evaluation is performed, so they won't
be evaluated, simplified,...
The 'groupname' allows to operate on
different selections at the same time.
Symaxx will try to move different selected
expressions with the same group name into
one selection, wherever possible.

Command:
SEL(Expr,Groupname)"}];
    my $C_SelRB3=['RB',[],{-text => 'Remove Selection',
			   -variable => \$F->{'APPLY_SEL_MODE'},
			   -value => 'UNSELECT',-balloon => 
'Apply this on an expression containing a selection 
to remove the selection. Any SEL(Expr,Groupname)
is just replaced by Expr, as long as the group name
matches or is "" (default).
Command: UNSEL(Expr,Groupname)'}];
    my $C_SelRB4=['RB',[],{-text => 'Pick out selected contents', 
			   -variable => \$F->{'APPLY_SEL_MODE'},
			   -value => 'PICK',-balloon =>
"This will isolate the contents of a selection.
It only works, if there is one single selection
with the wanted groupname within the expression, 
otherwise you'll get an error.
Command: SELPICK(Expr,Groupname)"}];
    my $C_SelRB5=['RB',[],{-text => 'All / default',
			   -variable => \$F->{'APPLY_SEL_GRP'},
			   -value => '*',-balloon => 
"Use this, if you only need a single selection group.
This will usually be the case.
The default selection is highlighted with an
orange box."}];
    my $C_SelRB6=['RB',[],{-text => 'Use group A',
			   -variable => \$F->{'APPLY_SEL_GRP'},
			   -value => 'A_R',-balloon =>
"The selection operation will operate on group A_R.
This selection is shown in red."}];
    my $C_SelRB7=['RB',[],{-text => 'Use group B',
			   -variable => \$F->{'APPLY_SEL_GRP'},
			   -value => 'B_G',-balloon =>
"Selection operation on group B_G,
highlighted in green"}];
    my $C_SelRB8=['RB',[],{-text => 'Use group C',
			   -variable => \$F->{'APPLY_SEL_GRP'},
			   -value => 'C_B',-balloon => 
"Selection operation on group C_B, blue.
If you need more groups, use the commands by hand.
_R, _B, _G, _Y, _P can be used as end of the group 
name to set the colour."}];

    my $SelUpdate=sub{	
	    my %P=@_;
	    my $Widget=$P{WIDGET} or die;
	    my $C=$P{CONTEXT} or die;
	    my %Context=%{$C};

	$F->{SEL_SELECT}='';
	$F->{SEL_UNSELECT}='';
	$F->{SEL_PICK}='';
	my $Thing=$Context{THING};  return unless defined $Thing;
	my $Index=$Context{INDEX}; return unless defined $Index;
	my $Path=$Context{PATH}; return unless defined $Path;

	my $Command=$F->{APPLY_SEL_MODE};
	if ($Command eq 'SELECT'){
	    $F->{SEL_SELECT}=$F->{APPLY_SEL_GRP};
	} elsif ($Command eq 'UNSELECT'){
	    $F->{SEL_UNSELECT}=$F->{APPLY_SEL_GRP};
	} elsif ($Command eq 'PICK'){
	    $F->{SEL_PICK}=$F->{APPLY_SEL_GRP};
	};
    };
    my $SelSetContext=sub{
	    $F->{APPLY_SEL_MODE}='';
	    $F->{APPLY_SEL_GRP}='*';
	if($F->{SEL_SELECT}){
	    $F->{APPLY_SEL_MODE}='SELECT';
	    $F->{APPLY_SEL_GRP}=$F->{SEL_SELECT};
	} elsif($F->{SEL_UNSELECT}){
	    $F->{APPLY_SEL_MODE}='UNSELECT';
	    $F->{APPLY_SEL_GRP}=$F->{SEL_UNSELECT};
	} elsif($F->{SEL_PICK}){
	    $F->{APPLY_SEL_MODE}='PICK';
	    $F->{APPLY_SEL_GRP}=$F->{SEL_PICK};
	};
    };


    return ['PAGE',[$C_SelRB1,$C_SelRB2,$C_SelRB3,$C_SelRB4,$C_SelRB5,$C_SelRB6,$C_SelRB7,$C_SelRB8],
			 {-label => 'Selection',
			  SETCODE => $SelSetContext,
			  UPDATECODE => $SelUpdate}];
}; #Generate_Command_SelectionPage

sub Generate_Command_UnitsPage{
    my %P=@_; my $F=$P{FLAGS} or die;
    my $C_UnitRB1=['RB',[],{-text => "Don't do anything (default)",
			    -variable => \$F->{'APPLY_UNIT_MODE'},
				-value => '',
				-balloon => "The 'Units' feature is not used in this evaluation."}];
	my $C_UnitRB2=['RB',[],{-text => 'Multiply with unit',
				-variable => \$F->{'APPLY_UNIT_MODE'},
				-value => 'MUL',-balloon => 'Multiplies the expression with the
unit. Just a way to insert a unit without having to type the name...'}];
	my $C_UnitRB3=['RB',[],{-text => 'Convert to unit',
				-variable => \$F->{'APPLY_UNIT_MODE'},
				-value => 'CONVERT',
-balloon => "Tries to convert an expression to a given unit.
- First divide the expression by the new unit
- Simplify
- Multiply again with the new unit without simplification
Command:ConvertUnit(Expr,TargetUnit)
*NOTE*: If you use that command, don't forget to
quote the target unit.
Example: ConvertUnit(Length,'CENTIMETER)
*NOTE2*: If you are using non-SI units, they will
be converted to SI at next evaluation.
To prevent this, apply a selection on the unit
directly after conversion."}]; 

	my $UnitLBSetContext=sub{	
    my %P=@_;
	    my $Widget=$P{WIDGET} or die;
	    my $C=$P{CONTEXT} or die;
	    my %Context=%{$C};

	    $F->{'APPLY_UNIT_MODE'}='';
	    $Widget->delete(0,'end');
	    my $Unit;foreach $Unit(@::AllUnits){
		$Widget->insert('end',$Unit);
	    };
	    my $Thing=$Context{'THING'}; return unless $Thing;
	    my $Index=$Context{'INDEX'}; return unless defined $Index;
	    my $ID=$Thing->QueryID;
	    my $WantedUnit;
	    if ($F->{UNIT_MULTIPLY}){
		$WantedUnit=$F->{UNIT_MULTIPLY};
		$F->{'APPLY_UNIT_MODE'}='MUL';
	    } elsif ($F->{UNIT_CONVERT}){
		$WantedUnit=$F->{UNIT_CONVERT};
		$F->{'APPLY_UNIT_MODE'}='CONVERT';
	    } else {return;}
	    $F->{UNIT}=$WantedUnit;
	    my @AllUnits=@::AllUnits;
	    my $Count; for ($Count=0;$Count<=$#AllUnits;$Count++){
		if ($AllUnits[$Count] eq $WantedUnit){
	#	    $Widget->configure(-state => 'normal');
		    $Widget->selectionSet($Count,$Count);
		    $Widget->yview($Count);
#		print $WantedUnit;print $AllUnits[$Count];print "\n";
		};
	    };
	}; # Sub UnitsListboxSetContext
	
	my $UnitLBUpdate=sub{		
	    $F->{UNIT_MULTIPLY}='';
	    $F->{UNIT_CONVERT}='';

	    # We have to assume a default unit once.
	    # The listbox cannot be without selection.
	    if ($F->{'APPLY_UNIT_MODE'} eq 'MUL'){
		if (!$F->{UNIT}){$F->{UNIT}='METERS'};
		$F->{UNIT_MULTIPLY}=$F->{UNIT};
		if (!$F->{'SKIPEVAL'}){
		    print "Turning off evaluation for this line, otherwise the new unit would probably be converted back.\n";
		    $F->{'SKIPEVAL'}=1;
		};
	    } elsif ($F->{'APPLY_UNIT_MODE'} eq 'CONVERT'){
		if (!$F->{UNIT}){$F->{UNIT}='METERS'};
		$F->{UNIT_CONVERT}=$F->{UNIT};
		if (!$F->{'SKIPEVAL'}){
		    print "Turning off evaluation for this line, otherwise the new unit would probably be converted back.\n";
		    $F->{'SKIPEVAL'}=1;
		};
	    } else {return;};
	};
	my $UnitLBClick=sub{ 
            my %P=@_;
	    my $Widget=$P{WIDGET} or die;

	    my $SelectionIndex=$Widget->curselection();
	    return unless defined $SelectionIndex;
	    my $ClickedString=$Widget->get($SelectionIndex);
	    return unless $ClickedString;
#	    print $ClickedString;
	    $F->{UNIT}=$ClickedString;
	};
	my $C_UnitListbox=['LB',[],{SETCODE => $UnitLBSetContext,
			 UPDATECODE => $UnitLBUpdate,
			   CLICKCODE => $UnitLBClick}];
return ['PAGE',[$C_UnitRB1,$C_UnitRB2,$C_UnitRB3,$C_UnitListbox],{-label => 'Units'}];
    }; # Generate_Line_UnitsPage
1;


	

