package MaximaController;
use strict;
use Symaxx2::ThingManager;
use POSIX ":sys_wait_h";
my $NL="\$\n"; # set to ';' to debug Maxima output 
my $MaybeBlanks='\s*';
my $StartOfLine='^';
my $EndOfLine="\$";
my $Errortoken1_unquoted="~:M";
my $Errortoken1=quotemeta($Errortoken1_unquoted);
my $Errortoken2=quotemeta('~%~M');
my $Errortoken3=quotemeta('~M'); # this appears for example, if error("Text",a,b) is used.
my $Errortoken4=quotemeta('~:@M'); # this appears for example, if integrate(1) is tried.
my $Errortoken5=quotemeta('~m'); # this appears for example, if loadfile("kaboom") is tried.
my $Space=' ';
sub Evaluate{
    my %P=@_;
    $::MainDirty=1;
    $::MaximaLog='';$::UIManager->SetErrorflag(0);
    # delete all files, so that we don't mix up data
    my @TempFiles=('MaximaOutput.mc','MaximaInput.mc','lispcode.mc','MaximaOutput','WatchdogLog','MaximaLog');
    my $File; foreach $File(@TempFiles){unlink "$::TempDir/$File";};
    
    # Make a list of all new commands (in ASCII), which have to be entered into Maxima.

    my @MaximaASCIIFile=();    
    if ($::UIManager->QueryFeature(FEATURE => 'FORMAT')){
	push @MaximaASCIIFile,"if errcatch(loadfile(\"$::CodeDir/MaximaBatch/format.o\"),TRUE)=[] then error(\"format.o not found. Run symaxx2 -compile\")$NL";
	push @MaximaASCIIFile,"if errcatch(loadfile(\"$::CodeDir/MaximaBatch/coeflist.o\"),TRUE)=[] then error(\"format.o not found. Run symaxx2 -compile\")$NL";
	push @MaximaASCIIFile,"format(3)$NL";
    };

    push @MaximaASCIIFile,@{&StartMaximaASCIIFile};
    my @EvalOrder;
    # if there are no new commands, then we prepare evaluation (also TeXOnly)
    my $EvalOrderResult=&GenerateEvalOrder; 
    if (!defined $EvalOrderResult){return 0;} # for example circular
    @EvalOrder=@{$EvalOrderResult};
    if ($#EvalOrder==-1){return 0;} # nothing to do
        
    # Generate Maxima Lisp save file for all those expressions, which have not been handled as ASCII
    push @MaximaASCIIFile,@{&GenerateLoadfileCommand};
    push @MaximaASCIIFile,@{&GenerateIncludeCommands};
    my $EvalData=&DoEval('EVALORDER' => \ @EvalOrder) or return 0;
    push @MaximaASCIIFile,@{$EvalData};
    
    &WriteMaximaASCIIFile('ASCII' => \@MaximaASCIIFile);
    
    # run Maxima
    if (!&RunMaximaToEvaluate){ 
	return 0;
    };
    
    #update all the objects, which have to be updated (either results or commands in input form)
    &ProcessMaximaOutput or return 0;
    # There may be some things around, which reference to nonexisting things.
    # Although they got evaluated as best as we could, they are not considered clean.
    $::ThingManager->MakeEvaldirtyIfNeeded;
    return 1;
};
sub ParseMaximaOutput{  
    my %P=@_;
    my $File=$P{FILE} or die;
    
    if (!(open InFile,"<$File")){
	$::MaximaLog.="Error: unable to open $File for read.\n";
	if (open M,"<$::TempDir/WatchdogLog"){
	    $::MaximaLog.="***Watchdog-log***\n";
	    my @Log=<M>;$::MaximaLog.=join('',@Log);
	    $::MaximaLog.="***End Watchdog-log***\n";
	};
	&ExamineLog; 
	$::UIManager->SetErrorflag(1); 
	return 0;
    };
    
    my $InLine; 
    my %AllVars=();
    my $Varname="";
    my $WholeText='';
    my $Record=0;
    while (defined($InLine=<InFile>)){	
	$InLine=~s/\n$//; # remove trailing carriage return, if any
	$InLine=~s/^\s+//; # remove leading whitespace
	$InLine=~s/\s+$//; # remove trailing whitespace
	if ($Record){
	    $WholeText.=" ".$InLine;
	} else {
	    # Bug fix: Mar 18,2001: Pavels and Michels save-output contains the line 
	    # (in-package "MAXIMA"
	    # the closing bracket is missing
	    # so here we check only for the opening bracket
	    if ($InLine=~/\(in\-package \"MAXIMA\"/i){$Record=1;}; # don't make it look like the package perl word
	}; # if record on
    }; # while file has data
    close InFile;
    $WholeText=~s/\n/ /g; # Carriage return to space
    $WholeText="($WholeText)"; # put into brackets, so that it is only one lisp object
    my $OK=&ExamineLog; # OK means that the script completed (the last line was evaluated)

    if ($WholeText=~/^\(\s*\)$/){
	if (!$OK){
	    $::MaximaLog.="*** ERROR: Maxima evaluation not finished ***\nMaxima did not create any output. Please check the log to find the reason.\n";
	    $::UIManager->SetErrorflag(1); 
	};
	return {};
    };
    my $AllDataAsLISP=LispList::NewFromLispString($WholeText);
#    $::MaximaLog.="***Reconstructed Maxima-output***\n".$AllDataAsLISP->QueryDebugString."\n\n\n";
    my $ObjectList=$AllDataAsLISP->QueryVariablesFromMaximaFile;
    
    return $ObjectList;
};
sub Maxima_LispToASCII{  
    $::MaximaLog="Converting to ASCII\n";$::UIManager->SetErrorflag(0);
    my %P=@_;
    my $Lispobject=$P{'LISPOBJECT'};
    GenerateMaximaInputFile('LIST'=>{'SYMAXX_ASCII'=>$Lispobject});
    open Out,">$::TempDir/MaximaInput" or die;
    print Out "loadfile(\"$::TempDir/lispcode.mc\");\n";
    print Out "stringout(\"$::TempDir/MaximaOutput\",SYMAXX_ASCII);\n";
    close Out;
    # Run Maxima
    my $Line;
    open MaximaLog,"$::Maxima < $::TempDir/MaximaInput|" or die;
    while (defined($Line=<MaximaLog>)){
#print $Line;
    }
    close MaximaLog;
    if (!open MaximaResult,"<$::TempDir/MaximaOutput"){
	$::MaximaLog.="Something went wrong with the Maxima evaluation. Maxima did not create an output file.\n";
 	$::UIManager->SetErrorflag(1); 
	return "Error";
    };
    my $MaximaResult='';
    while (defined($Line=<MaximaResult>)){chop $Line;$MaximaResult.=$Line;}
    close MaximaResult;
    # remove trailing ;
    if ($MaximaResult=~s/\;$//){
	$::MaximaLog.="Conversion succeeded.\n";
	return $MaximaResult;
    } else {
	$::MaximaLog.="Something went wrong with the Maxima evaluation. Here is the result: ".$MaximaResult." (end result)\n";
	$::UIManager->SetErrorflag(1); 
	return "Error";
    };
};

sub GenerateMaximaInputFile{
    my %P=@_;
    my $ObjectList=$P{'LIST'} or die;
    my @AllRefs=keys(%{$ObjectList});
    my %StringList; 
#    $::MaximaLog.="*** LISP Input file ***\n";

    my $Ref; foreach $Ref(@AllRefs){
	my $QuotedLispString=$ObjectList->{$Ref}->QueryLispCode('FORCEQUOTE' => 1) or die;
	# Example for $QuotedLispString: '((MPLUS) 1 1)
	$StringList{$Ref}=$QuotedLispString;
    };
    open Out,">$::TempDir/lispcode.mc" or die;
    print Out ";;; -*- Mode: LISP; package:maxima; syntax:common-lisp; -*-\n(in-package \"MAXIMA\") \n";
    foreach $Ref(@AllRefs){
	my $QuotedLispString=$StringList{$Ref}; 
#	$::MaximaLog.="$Ref >> $QuotedLispString\n";
	print Out "(DSKSETQ |\$$Ref\| $QuotedLispString) \n(ADD2LNC '|\$$Ref\| \$VALUES)\n";
    };
    close Out;
};
sub GenerateEvalOrder{
    my %P=@_;
    my %RemainingRefsToEvaluate=();
    my @EvalOrder=();
    # Make a list of all the references, that are needed
    # This means all references of the involved things and the involved things themselves
    
    my @q;
    @q=@{$::ThingManager->QueryListOfResultsForNextEvaluation;};	
    my $Ref;foreach $Ref(@q){
	$RemainingRefsToEvaluate{$Ref}=1;
#	print "In eval: $Ref\n";
    };
#    # Anything to do? (no need to create function definitions just for themselves
#    if (!keys %RemainingRefsToEvaluate){
#	$::MaximaLog.="Nothing to do!\n";
#	return [];
#    };
    
    # Some things are declared as 'function' definition (works with variables, too). 
    # They are put into the REF array before any evaluation takes place.
    # This must be done in TeX mode also, since QueryListOfResultsForNextEvaluation does not list definitions at all.
    my @FunDefsArray;
    @FunDefsArray=@{$::ThingManager->QueryDefinitions};

    # Check for definitions, which use references
    my @ErrorList=@{$::ThingManager->ListOfDefinitionsWhichUseReferences};
    if ($#ErrorList>-1){
    	$::MaximaLog.="***Error***\nA definition refers to another object.\nDefinitions are evaluated before 'ordinary' objects, so this does not make sense - the referred object won't be there.\nNote: I *could* think of uses for this construct, but it won't work reliably, believe me...\nList of faulty objects: ".join(",",@ErrorList)."\n\n";
	$::UIManager->SetErrorflag(1); 
    };
    
    # ___1___
    ### Find an order to evaluate everything ###
    # Go through the list of dirty expressions repeatedly.
    # For each of them: check, if all its references have been handled already (or it doesn't have any).
    # Move this over to the 'Evaluate' order and remove it from the 'Dirty' list.
    
    $::ThingManager->ClearCircular;
  RepeatUntilAllDoneLoop:while (%RemainingRefsToEvaluate){
      my @IDs=keys %RemainingRefsToEvaluate;
      my $ID; 
    DoEachIDLoop:foreach $ID(@IDs){
	my $Thing=$::ThingManager->QueryThing('ID'=>$ID) or die;
	my @ThingDependsOn=@{$Thing->QueryListOfReferenceIDs};
#	print "MC: $ID references ".join(" ",@AllItsReferences)."\n";
	my $Ref;
      AllItsReferencesLoop:foreach $Ref(@ThingDependsOn){

	  # @ThingDependsOn contains all the expressions, that have to be
	  # present at the time, when this thing is evaluated.
	  # All of them are or have been in %RemainingRefsToEvaluate.
	  # If one of them is still in %RemainingRefsToEvaluate, we can't
	  # evaluate this expression (because the referenced expression has
	  # not been handled). If on the other hand RemainingRefsToEvaluate does NOT
	  # contain the reference anymore, then it has been handled already.
	  if (!defined $RemainingRefsToEvaluate{$Ref}){
	      # Our expression $ID is referring to $Ref, but $Ref has already been evaluated.
	      # This is good, we continue with the dependencies.
	      next AllItsReferencesLoop;
	  };
	  
	  # There is a needed reference that is still present in the list
	  # of unhandled expressions. Therefore we can't evaluate this thing
	  # at this time. Try another thing.
	  next DoEachIDLoop;
      }; # AllItsReferencesLoop
	
	# At this point, none of the referenced expressions is still in
	# RemainingRefsToEvaluate - they will be evaluated earlier.
	# Now this thing can be evaluated and removed from RemainingRefsList.
	push @EvalOrder,$ID;
	delete $RemainingRefsToEvaluate{$ID};

	# We don't continue with the next unhandled expressions, 
	# but start over again with the first one.
	# If we ever reach the end of the list, all references have been 
	# tried but failed. There must be a circular reference then.
	next RepeatUntilAllDoneLoop;

	# Try the next ID. This loop is left via the 
	# next DoEachIDLoop statement. 
    }; # DoEachIDLoop
      
      # Ooookey, we have a circular reference. No reason to panic, though.
      # We turn on the circular flag on all remaining unhandled expressions,
      # turn on the warning light and bail out.
      $::MaximaLog.="Circular structure! Exiting.\n";
      foreach $ID(keys %RemainingRefsToEvaluate){
	  my $Thing=$::ThingManager->QueryThing('ID'=>$ID,'TOLERATEINVALID' => 1) or die;
	  $Thing->SetFlag(FLAGMODE => 'THING',FLAG => 'CIRCULAR', NEW => 1);
      };

      # We have to leave the loop now because of the circular condition.
      last RepeatUntilAllDoneLoop;
  };
    $::MaximaLog.="\nDefinitions: ".join(":",@FunDefsArray)."\n";
    $::MaximaLog.="\nNormal references: ".join(":",@EvalOrder)."\n";
    @EvalOrder=(@FunDefsArray,@EvalOrder);
    $::MaximaLog.="\nTotal evaluation order: ".join(":",@EvalOrder)."\n";
        
    return \@EvalOrder;
}; 
sub DoEval{
    my %P=@_;
    die unless $P{'EVALORDER'};
    my @EvalOrder=@{$P{'EVALORDER'}};
    my @MaximaASCIIFile=();
    my %LispobjectList=();
    my @SaveList=();
    my $ID;foreach $ID(@EvalOrder){
	my $Thing=$::ThingManager->QueryThing('ID'=>$ID);
	$Thing->ClearErrors; 

	my @Commands=@{$Thing->QueryCommands or die};
	    
	    # Process each command level
	    my $Index;for ($Index=0;$Index<=$#Commands;$Index++){
		my $Command=$Commands[$Index] or die;
		
		my @PathList=@{$Thing->QuerySortedPaths('INDEX' => $Index,'LONGEST_FIRST' => 1)};
		my $Path; foreach $Path(@PathList){	
		    my %Data=(INDEX => $Index, ID => $ID, THING => $Thing, PATH => $Path,COMMANDCOUNT => $#Commands);
#			print "Generating Eval-Code: ID $ID Index $Index Path $Path\n";
		    my %WhatToDo=&ControlLogic(%Data);
		    my $EvalIt=$WhatToDo{EVALIT};
		    my $RecallIt=$WhatToDo{RECALLIT};
		    my $SaveIt=$WhatToDo{SAVEIT};
		    my $StartNewEval=$WhatToDo{STARTNEWEVAL};

		    if ($StartNewEval){
			push @MaximaASCIIFile,&StartNewEvaluation(%Data);
		    }; # if first command
		    		    
		    if ($EvalIt){
			my $Path2='';if ($Path ne ''){$Path2=','.$Path;};
			push @MaximaASCIIFile, "TTYOFF:FALSE$NL";
			push @MaximaASCIIFile, "print(\"[SYMAXX-STARTEVAL] -$ID- -$Index- -$Path-\")$NL";			
			push @MaximaASCIIFile,"\n/*Now substituting: ID $ID Index $Index Path $Path*/\n";
 
			my $SubstVarName=&GetSubstVarname('ID' => $ID,'INDEX' => $Index,'PATH' => $Path);
			$Data{SUBSTVARNAME}=$SubstVarName;
			# Make the lisp object for this substitution available
			$LispobjectList{$SubstVarName}=$Command->{$Path}->{'LISPOBJECT'} or die $Path;   
			# Replace % in the current substitution by the expression, that it is going to replace.
			# This will get all references by % in place without actually evaluating 
			
			push @MaximaASCIIFile,"/*Now replacing % within the command with the expression that the command is applied to*/\n";
			push @MaximaASCIIFile,"$SubstVarName:subst(INPART(REF[$ID]$Path2),'%,$SubstVarName)$NL";
			push @MaximaASCIIFile,"ErrorReporter()$NL";

			push @MaximaASCIIFile,&ApplyFeatures(%Data);


			# Then move the substitution command into place.
			push @MaximaASCIIFile,"/*Now moving the substitution command into place*/\n";
			# We wrap the new command into SYMAXX_SUBSTITUTED(). This is necessary to replace
			# the original expression (with length 1) with a new expression, which must also
			# have length 1. ???
			# push @MaximaASCIIFile,"REF[$ID]:SUBSTINPART(SYMAXX_SUBSTITUTED($SubstVarName),REF[$ID]$Path2)$NL";		
			push @MaximaASCIIFile,"simp:off$NL
REF[$ID]:SUBSTINPART($SubstVarName,REF[$ID]$Path2)$NL";
			push @MaximaASCIIFile,"ErrorReporter()$NL";

			# Apply standard substitution rules to turn SEL(a) into SEL(a,"")
			push @MaximaASCIIFile,"REF[$ID]:apply1(REF[$ID],SYMAXX_DefaultizeProtect)$NL";
			

			if ($Path eq ''){
			    push @MaximaASCIIFile,&MoveAllRefsInPlace(%Data);
			    push @MaximaASCIIFile,&AddEvalStatement(%Data);
			    push @MaximaASCIIFile,&EndProtected(%Data);
			    push @MaximaASCIIFile,&ApplySelSimp(%Data);
			}; # if root path 
			if ($SaveIt){
			    
			    if ($Index == $#Commands){
				# We'll put all E-labels into a list together with the last result.
				# This preserves compatibility with original Maxima language, only the result appears in a different way.
				push @MaximaASCIIFile,&HandleELabels(%Data);			
			    }; # if last line

			    push @MaximaASCIIFile,&ApplyDispform(%Data);

			    # Save the result

			    # Arrays are very complicated to handle in the save file, so we move the result to a normal variable.
			    # Totaldisrep has to be done, because CRE-expressions will crash Symaxx (for example RAT(3))
			    my $SaveVarname=&GetSrcVarname('ID' => $ID,'INDEX' => $Index);
			    $SaveVarname.='SAVE'; # This addition is only a comment
			    push @SaveList,$SaveVarname;			    
			    push @MaximaASCIIFile,"block([SIMP:OFF],$SaveVarname:totaldisrep(REF[$ID]))$NL";
			    # Push the result into our %TH(x)-substitute history (command or recall)
			    push @MaximaASCIIFile,"block([SIMP:OFF],SYMAXX_PUSH_INTO_HISTORY(REF[$ID]))$NL";
			    push @MaximaASCIIFile,"ErrorReporter()$NL";

			    # End save
			}; # if SaveIt
		    }; # if EvalIt
		    if ($RecallIt){
			my $Path='';
			push @MaximaASCIIFile, "print(\"[SYMAXX-STARTEVAL] -$ID- -$Index- -$Path-\")$NL";
			push @MaximaASCIIFile,"\n/*Now recalling: ID $ID Index $Index Path $Path*/\n"; 
			# It is possible, that the result is undefined, for whatever reason.
			# We don't load it, and REF[42] remains REF[42] after evaluation.
			my $Lispresult=$Command->{$Path}->{'LISPOBJECT_RESULT'};
			if ($Lispresult){
			    # Load the original command into REF[$ID]
			    my $SrcVarname=&GetSrcVarname('ID' => $ID,'INDEX' => $#Commands);
			    $SrcVarname.='RESULT'; # this only serves as comment.
			    $LispobjectList{$SrcVarname}=$Lispresult;
			    push @MaximaASCIIFile,"REF[$ID]:$SrcVarname$NL";
			};
		    }; # if RecallIt
		}; # for path
	    }; # For index
	push @MaximaASCIIFile, "print\(\"[SYMAXX-ENDEVAL]\"\)$NL";
	push @MaximaASCIIFile,"print(\"Evaluation of ID $ID succeeded.\")$NL";
    }; # foreach ID
    if ($#SaveList !=-1){
	push @MaximaASCIIFile,"save(\"$::TempDir/MaximaOutput.mc\",".join(',',@SaveList).")".$NL;
    } else {
	$::MaximaLog.="*** Nothing to do! ***\n";
	return undef;
    };
#    print join(",",%LispobjectList)."\n";
    &GenerateMaximaInputFile('LIST' => \ %LispobjectList);
    return \@MaximaASCIIFile;
};
	
sub Watchdog{
    my %P=@_;
    if ($::WatchdogEnabled){
	my $OnOff=$P{'ONOFF'};die unless defined $OnOff;
	if (!$OnOff){
	    alarm(0);
	    if ($::WatchdogPID){kill 'SIGKILL',$::WatchdogPID;}
	} else {
	    my $PID;if (!($PID=fork)){
		exec("$::CodeDir/Symaxx2/Watchdog $::TempDir $::MaximaPID"); 
	    };
	    $::WatchdogPID=$PID;
	};
    };
};
sub GetFilenameWithoutExtension{
    my $n=$::Filename;    
    if ($n=~/^(.*)\.mth$/){$n=$1;};# remove extension
    return $n;
};
sub StartMaximaASCIIFile{
    my $FileNameNoExt=&GetFilenameWithoutExtension;
    my $Gnuplot=q{gnuplot(MyFunc,MyVar,range,number_ticks,file):=BLOCK([NUMER:TRUE,DISPLAY2D:FALSE,EPSFile:SCONCAT(SymaxxWorkDir,EV(file)),resuneval,res,x],MyVar:EV(MyVar),WITH_STDOUT("TempDir",FOR x FROM EV(range[1]) STEP EV(range[2]-range[1])/EV(number_ticks) THRU EV(range[2]) DO BLOCK(x:ev(x,numer),res:EV(ev(MyFunc),ev(MyVar)=x,EVAL,NUMER),IF NUMBERP(x) AND NUMBERP(res) THEN PRINT(x,res) ELSE ERROR("ERROR: non-numeric in gnuplot var:",x," func: ",res))),a:SCONCAT(" echo \"set data style lines\; set title '",MyFunc,"'\; set out ","'",EPSFile,"'\; ","set term postscript eps\;plot 'TempDir' \" | gnuplot"),SYSTEM(a),"DONE")}.$NL;
    $Gnuplot=~s/TempDir/$::TempDir\/Plot.data/g;

    return [
	    # We can't simply turn off labels, because the E-labels of 
	    # LDISPLAY are not generated anymore with NOLABELS:TRUE.
	    # Therefore we change the labels to SYMAXX_C and D.
	    # The response for this line has to be killed, anyway.
	    "INCHAR:SYMAXX_C$NL",
	    "OUTCHAR:SYMAXX_D$NL",
	    "KILL(CLABELS,C1,C2,C3)$NL",
	    "KILL(DLABELS,D1,D2,D3)$NL",	    
	    # Set Symaxx predefined variables
	    "SymaxxWorkDir:\"$::WorkDir/\"$NL",
	    "SymaxxTempDir:\"$::TempDir/\"$NL",
	    "SymaxxFileName:\"$FileNameNoExt\"$NL",
	    $Gnuplot,                    # Gnuplot function
	    "SYMAXX_SIMP:TRUE$NL",           # corresponds to Maxima-default

	    # Set include-path for batch command
	    "FILE_SEARCH_MAXIMA:append(FILE_SEARCH_MAXIMA,[\"$::CodeDir/MaximaBatch/\"])$NL"
	    ];
					   
}; 
sub GenerateLoadfileCommand{
    # create a statement, that loads the lispfile that Symaxx just made.
    return ["loadfile(\"$::TempDir/lispcode.mc\")$NL",	 
	    # Load Symaxx-definitions - must be done after loading lispcode.mc
	    'batch("SymaxxInclude.mc")'.$NL
	    ];
};
sub GenerateIncludeCommands{
    # Units package (newton, meter etc)
    my @Code=(); 
    
    if ($::UIManager->QueryFeature(FEATURE => 'UNITS')){
	push @Code,"batch(\"$::CodeDir/MaximaBatch/SymaxxUnits.mc\")$NL";
    };
    if ($::UIManager->QueryFeature(FEATURE => 'GNUPLOT')){
	push @Code,"batch(\"$::CodeDir/MaximaBatch/SymaxxGnuplot.mc\")$NL";
    };
    if ($::UIManager->QueryFeature(FEATURE => 'NUMERICAL')){
	push @Code,"batch(\"$::CodeDir/MaximaBatch/SymaxxNumericalFunctions.mc\")$NL";
    };
    return \@Code;
};
sub CreateEvalArgs{
    my %P=@_;
    my $Thing=$P{'THING'} or die;
    my $RefsInEval=$Thing->QueryRefsInEval;
    my $EvArgs='';

    die;
    if ($Thing->QueryFlag(FLAGMODE => 'COMMAND', FLAG => 'EVAL')){$EvArgs.=',eval';}
    if ($Thing->QueryFlag(FLAGMODE => 'COMMAND', FLAG => 'NUMER')){$EvArgs.=',numer';}
    return $EvArgs;
};	

sub CreateTeXCode{ 
    my %P=@_;
    my $Thing=$P{'THING'} or die;
    my @MaximaASCIIFile=();
    $Thing->UpdateTexAtNextRedraw;
    my $ID=$P{'ID'}; die unless defined $ID;
    my $TotalPicName=MySubs::PictureTotalName('ID' => $ID,'IDENTIFIER' => 'TEXPICRES');
    my $TexFilename=$::TempDir."/$TotalPicName.tex";
    open NewFile,">$TexFilename" or die;
    print NewFile "\%\&\n"; # tex header
    print NewFile "\\hsize=100in\n"; # make paper wide
    print NewFile "\\nopagenumbers\n"; # turn off page numbers
    close NewFile;
    push @MaximaASCIIFile,"tex(REF[$ID],\"$TexFilename\")$NL";
    return \ @MaximaASCIIFile;
};  
sub WriteMaximaASCIIFile{ 
    my %P=@_;
    my @MaximaASCIIFile=@{$P{'ASCII'}};
#    $::MaximaLog.="\n*** ASCII input file to Maxima: ***\n";
    open OutFile,">$::TempDir/MaximaInput" or die "ERROR: Failed to open $::TempDir/MaximaInput for write";
    my $Line;
    push @MaximaASCIIFile,"print(\"[SYMAXX-EVALSUCCESS]\")$NL";
#    $::MaximaLog.="\n\n---Begin ASCII-input---\n";
    foreach $Line(@MaximaASCIIFile){
	print OutFile $Line;
#	$::MaximaLog.=$Line;
    };
#    $::MaximaLog.="---End ASCII-input---\n\n";
    close OutFile;
};
sub RunMaximaToEvaluate{
    # 25 years of glorious Unix history.
    # And then something like this.
    if ($::MaximaPID=fork){
	$SIG{'ALRM'}=\&WatchdogAlarm;
	$SIG{'INT'}='IGNORE'; # ignore INT-message from Watchdog or console
	alarm($::WatchdogAlertTime);
	waitpid($::MaximaPID,0);
	&Watchdog('ONOFF' => 0);
	$SIG{'INT'}= \ (&UIManager::Quit_br); # catch Ctrl-C
	return 1;
    } else {
	exec("$::Maxima < $::TempDir/MaximaInput > $::TempDir/MaximaLog");
    };
};
sub ProcessMaximaOutput{
    my %P=@_;
    my $ret=&ParseMaximaOutput('FILE' => "$::TempDir/MaximaOutput.mc") or return 0;
    my %NewObjectList=%{$ret;};
    my $RefXX; my $LispObject; 
    while (($RefXX,$LispObject)=each(%NewObjectList)){
	my $FieldInThing='';
	$RefXX=~/^\s*\$SYMAXX_CMDSRC_ID(\d+)_INDEX(\d+)SAVE\s*$/ or die $RefXX;
	my $ID=$1; my $Index=$2;
	my $Thing=$::ThingManager->QueryThing('ID' => $ID) or die;
	$Thing->SetLispResult('INDEX' => $Index,'NEW' => $LispObject);
    }; # while new object 
    return 1;
};

sub WatchdogAlarm{
    &Watchdog('ONOFF' => 1);
};
sub ExamineLog{
    my $OK=0;
    if (open Log,"<$::TempDir/MaximaLog"){
	my @Data_=<Log>;close Log;
	my $Linecount=0; 
	my $Current_Eval_Position;
	my $ErrorsAreValid;
	my $LogLine; my $LastLogLine='';

	# merge everything between [MAXIMA-ERROR-START] and
	# [MAXIMA-ERROR-END] into one line
	my @Data; my $ErrorString='';
	foreach $LogLine(@Data_){
	    chop $LogLine;
	    if (($LogLine=~/\[MAXIMA-ERROR-START\]/) ||
		($LogLine=~/\[MAXIMA-ERROR-END\]/)){
		if ($ErrorString ne ''){$ErrorString.=' ';}
		$ErrorString.=$LogLine;   
		if ($ErrorString =~/\[MAXIMA-ERROR-END\]/){
		    push @Data,$ErrorString; 
		    $ErrorString='';
		};
	    } else {
		# if neither START nor END in this line
		if ($ErrorString eq ''){
		    # no error? Append to log
		    push @Data,$LogLine;	
		} else {
		    # Error present? append to error.
		    if ($ErrorString ne ''){$ErrorString.=' ';}
		    $ErrorString.=$LogLine;  
		};
	    }; # if neither START nor STOP
	}; # foreach line
	if ($ErrorString ne ''){	
	    push @Data,$ErrorString;	
	};

	LogLineLoop:foreach $LogLine(@Data){
	    $Linecount++;
	    
	    # Handle log lines, which give general information (start, end, address of current
	    # evaluation)
	    my $Result=&CheckLoglineForStartEval($LogLine);
	    if (defined $Result){
		$Current_Eval_Position=$Result; 
		$ErrorsAreValid=1; $LastLogLine='';next LogLineLoop;
	    };

	    $Result=&CheckLoglineForEndEval($LogLine);
	    if ($Result){
		my $ID='unknown'; 
		if (defined($Current_Eval_Position->{ID})){
		    $ID=$Current_Eval_Position->{ID};	
		};
		$::MaximaLog.="Evaluation of ID $ID - OK\n";
		undef $Current_Eval_Position;$LastLogLine='';next LogLineLoop;
	    };
	    
	    $Result=&CheckLoglineForEvalSuccess($LogLine);
	    if ($Result){
		$::MaximaLog.="Evaluation completed (script finished).\n";
		$OK=1;$LastLogLine='';next LogLineLoop;
	    };
	    $Result=&CheckLoglineForSymaxxError($LogLine);
	    if ($Result){
		my $ID='unknown'; 
		if (defined($Current_Eval_Position->{ID})){
		    $ID=$Current_Eval_Position->{ID};	
		    my $Index=$Current_Eval_Position->{INDEX}; 
		    my $Path=$Current_Eval_Position->{PATH};

		    # Every evaluation can only set one error. Afterwards $ErrorsAreValid 
		    # is cleared.
		    if ($ErrorsAreValid && defined($Index) && defined($Path)){
			my $Errorthing=$::ThingManager->QueryThing(ID => $ID) or die;
			$Errorthing->SetError(INDEX => $Index,PATH => $Path,ERROR => $Result);
			undef $ErrorsAreValid;
		    }; # if exact location of error is known
		};
		$::MaximaLog.="Error in ID $ID: $Result at log line $Linecount\n";	
		$::UIManager->SetErrorflag(1); 
		$LastLogLine='';next LogLineLoop;
	    };

	    if (0){
		# Look out for non-Symaxx generated error message
		# (i.e. Division by zero
		# -- an error. Quitting. To debug this try...
		# It doesn't hurt, if we register the same error several times.
		
		# ***WRONG***. It DOES hurt. It's annoying.

		if (&CheckLoglineForGeneralError($LogLine)){
		    my $ID='unknown'; 
		    if (defined($Current_Eval_Position->{ID})){
			$ID=$Current_Eval_Position->{ID};	
			my $Index=$Current_Eval_Position->{INDEX}; 
			my $Path=$Current_Eval_Position->{PATH};
			
			if (defined($Index) && defined($Path)){
			    my $Errorthing=$::ThingManager->QueryThing(ID => $ID) or die;
			    $Errorthing->SetError(INDEX => $Index,PATH => $Path,ERROR => "$LastLogLine");
			}; # if exact location of error is known
		    };
		    $::MaximaLog.="Error in ID $ID: $LastLogLine at log line $Linecount\n";	
		    $::UIManager->SetErrorflag(1); 
		    
		    next LogLineLoop;
		};
	    };

	    # It can happen, that Maxima crashes so badly, that the error reporter does not work anymore.
	    # For example, solve([1,2],[a,b]) seems to do just that.
	    # Look for indications, that this has happened and try to find the error message.
	    if (&CheckLoglineForFatalError($LogLine)){
		my $ID='unknown'; 
		if (defined($Current_Eval_Position->{ID})){
		    $ID=$Current_Eval_Position->{ID};	
		    my $Index=$Current_Eval_Position->{INDEX}; 
		    my $Path=$Current_Eval_Position->{PATH};
		    
		    if (defined($Index) && defined($Path)){
			my $Errorthing=$::ThingManager->QueryThing(ID => $ID) or die;
			$Errorthing->SetError(INDEX => $Index,PATH => $Path,ERROR => "<FATAL>");
		    }; # if exact location of error is known
		};
		$::MaximaLog.="Error in ID $ID: $LastLogLine at log line $Linecount\n";	
		$::UIManager->SetErrorflag(1); 
		
		# At this point it does not make sense to try to further process the log.
		last LogLineLoop;
	    };

#	    if (0){
#		if ($LogLine=~/Incorrect syntax/){
#	   	$::MaximaLog.="Maxima-Syntax-error";
#		if (defined($Eval_ID)){$::MaximaLog.="in ID $Eval_ID";};
#		$::MaximaLog.="\n\"$LogLine\"\n";
#		$::UIManager->SetErrorflag(1);	
#		$FirstError=0;
#	    } elsif ($LogLine=~/Error: Bind stack overflow/){
#		$::MaximaLog.="Maxima-error:Bind stack overflow";
#		if (defined($Eval_ID)){$::MaximaLog.="in ID $Eval_ID";};
#		$::MaximaLog.="\n\"$LogLine\"\n";
#		$::UIManager->SetErrorflag(1); 
#		$FirstError=0;
#	    } elsif ($LogLine=~/Error: (.*)$/){
#		$::MaximaLog.="Maxima-error: $1";
#		if (defined($Eval_ID)){$::MaximaLog.="in ID $Eval_ID";};
#		$::MaximaLog.="\n\"$LogLine\"\n";
#		
#		$::UIManager->SetErrorflag(1);	
#		$FirstError=0;
#	    };
#	    }; # if 0
     
	    $LastLogLine=$LogLine;
	}; # foreach log line
	
	my $TotalLogData=join("\n",@Data);
#	if($TotalLogData=~/an error.*Quitting.*To debug this try/i || $TotalLogData=~/Incorrect syntax:/){

	# append full log, if an error has occured
#	if (!$FirstError){
#	    $::MaximaLog.="*** Error *** (log follows)\n";
#	    $::UIManager->SetErrorflag(1); 
	    $::MaximaLog.=$TotalLogData."*** End log ***\n";
	    # in this case Maxima generates output and some results
	    # may be OK
#	}
    } else {  
	$::MaximaLog.= "***Oops...Unable to open Maxima-log ?!***";
    }; # if log open succeeded
    return $OK;
};

sub GenerateModifCode{ 
    my %P=@_;
    my $What=$P{'WHAT'}; die unless $What;
    my $ID=$P{'ID'}; die unless defined $ID;
    
    my $Thing=$::ThingManager->QueryThing('ID' => $ID) or die;
    my $Modifs=$Thing->{'MODIF_'.$What};
    my $ReturnCode='';
    if ($Modifs){ # Array
	my %CopyOfModifs=%{$Modifs};
	my $LongestPath=''; 
	my @Paths=keys(%CopyOfModifs);
	while ($#Paths>=0){
	    my $MaxLength=-1;
	    my $Path; foreach $Path(@Paths){
		my $Length;
		if ($Path eq ''){$Length=0} else {
		    my @PathArray=split(/,/,$Path);
		    $Length=$#PathArray;
		};
		if ($Length > $MaxLength){
		    $MaxLength=$Length; $LongestPath=$Path;
		};
	    }; # foreach path
	    my $Modifier=$CopyOfModifs{$LongestPath};
	    my $PartString=$LongestPath; if ($PartString ne ''){$PartString=",$PartString";} # PartString can be empty for toplevel expression 
	    
	    # Retrieve selected element and put it into % 
	    $ReturnCode.="%:INPART(REFRES$ID$PartString)$NL";
	    $LongestPath=~s/\,/_/g; # replace , by _
	    $ReturnCode.="SYMAXXMODIF:subst(%,'%,SYMAXXMODIF_ID".$ID."_".$What."_".$LongestPath.")$NL";
	    $ReturnCode.="REFRES$ID:SUBSTINPART(SYMAXXMODIF,REFRES$ID$PartString)$NL"."simp:on$NL";
	    delete $CopyOfModifs{$LongestPath};
	    @Paths=keys(%CopyOfModifs);
	};
    };
    return $ReturnCode;
};
sub Maxima_ASCII_to_LISP{
    my %P=@_;
    my $ASCII=$P{'ASCII'}; die unless defined $ASCII;
    my @F=(); # ASCII file to Maxima
    $::MaximaLog='';$::UIManager->SetErrorflag(0);
    if ($::UIManager->QueryFeature(FEATURE => 'FORMAT')){
	push @F,"if errcatch(loadfile(\"$::CodeDir/MaximaBatch/format.o\"),TRUE)=[] then error(\"format.o not found. Run symaxx2 -compile\")$NL";
	push @F,"if errcatch(loadfile(\"$::CodeDir/MaximaBatch/coeflist.o\"),TRUE)=[] then error(\"format.o not found. Run symaxx2 -compile\")$NL";
    };

    push @F,"simp:off;SYMAXX_RESULT:'($ASCII)$NL";
    push @F,"string(SYMAXX_RESULT)$NL"; # Some strange commands may cause a crash when they are evaluated. We cause the crash here already => no save => no new data for Symaxx
    push @F,"save(\"$::TempDir/MaximaOutputNewline.mc\",SYMAXX_RESULT)$NL";
    &WriteMaximaASCIIFile('ASCII' => \@F);
    # run Maxima
    if (!&RunMaximaToEvaluate){ 
	&ExamineLog; # if there is any
	$::UIManager->SetErrorflag(1); 
	return 0;
    };
    my $LispStuff=&ParseMaximaOutput('FILE' => "$::TempDir/MaximaOutputNewline.mc");# or die;
    if (!$LispStuff->{'$SYMAXX_RESULT'}){
	$::MaximaLog="*** Error ***\nThe conversion of '$ASCII' into Maxima has failed.\n\n\n".$::MaximaLog;
	$::UIManager->SetErrorflag(1); 
	return undef;
    };
    
    my $LispObject=$LispStuff->{'$SYMAXX_RESULT'} or die;
    return $LispObject;
};

sub GetSubstVarname{
    my %P=@_;
    die unless defined $P{'ID'};
    die unless defined $P{'INDEX'};
    my $Path=$P{'PATH'}; die unless defined $Path;
    $Path=~s/\,/\_/g; # replace , by _ in path
    my $Varname='SYMAXX_CMDSUBST_';
    $Varname.='ID'.$P{'ID'};
    $Varname.='_INDEX'.$P{'INDEX'};
    $Varname.='_PATH'.$Path;
    return $Varname;
};
sub GetSrcVarname{
    my %P=@_;
    die unless defined $P{'ID'};
    die unless defined $P{'INDEX'};
    return 'SYMAXX_CMDSRC_ID'.$P{'ID'}.'_INDEX'.$P{'INDEX'};
    };
1;
sub AddEvalStatement{
    my %P=@_;
    my $ID=$P{ID} or die;
    my $Thing=$P{THING} or die;
    my $Index=$P{INDEX}; die unless defined $Index;
    my @Code=();

    # ################################################################
    # ### Create code for the different Eval-args                  ###
    # ################################################################

    my $EvArgs='';
    # We *must* turn on TTY during evaluation.
    # Otherwise, the questions 'is x positive or negative
    # will not be heard!
    my $PreEval='TTYOFF:FALSE;';
    my $PostEval='TTYOFF:TRUE;';
    
    my $DoNotSimp=$Thing->QueryFlag
	(FLAGMODE => 'LINE',
	 INDEX => $Index,
	 FLAG => 'DONOTSIMP');
    if ($DoNotSimp){
	$PreEval.="SIMP:OFF$NL";
    };
    
    my $Numer=$Thing->QueryFlag
	(FLAGMODE => 'LINE',
	 INDEX => $Index,
	 FLAG => 'NUMER');
    if ($Numer){$EvArgs.=',NUMER'};
    
    my $EvalFlag=$Thing->QueryFlag
	(FLAGMODE => 'LINE',
	 INDEX => $Index,
	 FLAG => 'EVAL');
    if ($EvalFlag){$EvArgs.=',EVAL'};
    
    # ################################################################
    # ### Create code for the references                           ###
    # ################################################################

    my $RefsInEval=$Thing->QueryFlag
	(FLAGMODE => 'LINE',
	 INDEX => $Index,
	 FLAG => 'REFSINEVAL');
    if ($RefsInEval){
	my $EvalRef; foreach $EvalRef(@{$RefsInEval}){
	    # check, if reference is valid
	    my $ReferencedThing=$::ThingManager->QueryThing('ID' => $EvalRef,'TOLERATEINVALID' => 1);
	    if (!$ReferencedThing){
		my $OwnID=$Thing->QueryID;
		$::MaximaLog.="*** ERROR: Invalid reference***\nThe list of references for the ev()-statement of ID $OwnID includes ID $EvalRef, which does not exist.\n";
		$::UIManager->SetErrorflag(1); 
	    };
	    $EvArgs.=",REF[$EvalRef]";
	};
    };	
    
    my $SkipEval=$Thing->QueryFlag
	(FLAGMODE => 'LINE',
	 INDEX => $Index,
	 FLAG => 'SKIPEVAL');
    # Here comes the final EV-statement (if wanted and needed)		    
    if (!$SkipEval){ 
	if ($::UIManager->QueryFeature(FEATURE => 'SELECTION')){
	    push @Code,"REF[$ID]:apply1(REF[$ID],SYMAXX_StartProtect)$NL";
	};
	push @Code,"simp:TRUE$NL";
	push @Code,$PreEval;
	push @Code,"REF[$ID]:ev(REF[$ID]$EvArgs)$NL";			
	push @Code,$PostEval;
    };	
    return @Code;
};
sub MoveAllRefsInPlace{
    my %P=@_;
    my $ID=$P{ID} or die;
    my $Thing=$P{THING} or die;
    my $Index=$P{INDEX}; die unless defined $Index;
    my @Code=();
    # Now get the contents of all REF[x]-expressions into place. 
    push @Code,"REF[$ID]:apply1(ev(REF[$ID],NOEVAL),SYMAXXRefEval)$NL";
    return @Code;
};	
sub EndProtected{
    my %P=@_;
    my $ID=$P{ID} or die;
    my $Thing=$P{THING} or die;
    my $Index=$P{INDEX}; die unless defined $Index;
    my @Code=();
    if ($::UIManager->QueryFeature(FEATURE => 'SELECTION')){
	push @Code,"SYMAXX_SIMP:simp$NL";
	push @Code,"REF[$ID]:apply1(REF[$ID],SYMAXX_EndProtect)$NL";
	push @Code,"ErrorReporter()$NL";
    };
    return @Code;
};
sub ApplySelSimp{
    my %P=@_;
    my $ID=$P{ID} or die;
    my $Thing=$P{THING} or die;
    my $Index=$P{INDEX}; die unless defined $Index;
    my @Code=();
    if ($::UIManager->QueryFeature('FEATURE' => 'SELECTION')){
	my $ApplySelSimp=$Thing->QueryFlag
	    (FLAGMODE => 'LINE',
	     INDEX => $Index,
	     FLAG => 'SEL_SELSIMP');
	if ($ApplySelSimp){
	    push @Code,"REF[$ID]:apply1(REF[$ID],SelSimp)$NL";
	    push @Code,"ErrorReporter()$NL";
	};
    };
    return @Code;
};
sub HandleELabels{	
    my %P=@_;
    my $ID=$P{ID} or die;
    my $Thing=$P{THING} or die;
    my $Index=$P{INDEX}; die unless defined $Index;
    my @Code=();    
    push @Code, "if (length(labels(E)))#0 then REF[$ID]:append([REF[$ID]],[MAP(LAMBDA([U],U=ev(U)),labels(E))])$NL";	
    return @Code;
};
sub ApplyFeatures{
    my %P=@_;
    my $ID=$P{ID} or die;
    my $Thing=$P{THING} or die;
    my $Index=$P{INDEX}; die unless defined $Index;
    my $Path=$P{PATH}; die unless defined $Path;
    my $SubstVarName=$P{SUBSTVARNAME} or die;
    my @Code=();    
    
    push @Code,"SYMAXX_ID:$ID$NL";
    push @Code,"SYMAXX_INDEX:$Index$NL";
    
    if ($::UIManager->QueryFeature('FEATURE' => 'SELECTION')){
	# Apply protect / unprotect / pick, when selected
	# through the popup window (just writing
	# SEL(expr) does not require the next lines)
	my $Mode;
	foreach $Mode('UNSELECT','SELECT','PICK'){
	    my $Group=$Thing->QueryFlag
		(FLAGMODE => 'COMMAND',
		 INDEX => $Index,
		 PATH => $Path,
		 FLAG => 'SEL_'.$Mode);
	    print $Mode,$Group."\n";
	    if ($Group){
		my $Arg=",$Group";
		# * => ''
		if ($Group eq '*'){$Arg='';} 
		my $Command;
		if ($Mode eq 'UNSELECT'){
		    $Command='UNSEL';
		} elsif ($Mode eq 'SELECT'){
		    $Command='SEL';
		} elsif ($Mode eq 'PICK'){
		    $Command='SELPICK';
		} else {die;};
		push @Code,"/* $Mode is selected as a command attribute - wrap cmd. subst. in $Command*/\n";
		push @Code,"$SubstVarName:funmake($Command,[$SubstVarName$Arg])$NL";
		# Error will happen when executing.
		# There is not much that can go wrong with 'funmake'.

	    };
	}; # foreach Mode
    };
    if ($::UIManager->QueryFeature('FEATURE' => 'UNITS')){
	my $ConvertToUnit=$Thing->QueryFlag
	    (FLAGMODE => 'COMMAND',
	     INDEX => $Index,
	     PATH => $Path,
	     FLAG => 'UNIT_CONVERT');
	if ($ConvertToUnit){
	    push @Code,"/* Feature UNITS - converting to $ConvertToUnit.*/\n";
	    push @Code,"$SubstVarName:block([SIMP:ON],ConvertUnit($SubstVarName,'($ConvertToUnit)))$NL";
	};
	my $MulWithUnit=$Thing->QueryFlag
	    (FLAGMODE => 'COMMAND',
	     INDEX => $Index,
	     PATH => $Path,
	     FLAG => 'UNIT_MULTIPLY');
	if ($MulWithUnit){
	    push @Code,"/* Feature UNITS - multiplying with $MulWithUnit.*/\n";
	    push @Code,"block([SIMP:OFF],$SubstVarName:'($MulWithUnit)*$SubstVarName)$NL";
	};
    }; # if UNITS-Feature
    if ($::UIManager->QueryFeature(FEATURE => 'GNUPLOT')){
	if ($Thing->QueryFlag
	    (FLAGMODE => 'LINE',
	     INDEX => $Index,
	     FLAG => 'GNUP_CREATEDATA')){
	    my $Indep=$Thing->QueryFlag
		     (FLAGMODE => 'LINE',
		      INDEX => $Index,
		      FLAG => 'GNUP_INDEP');
	    my $Min=$Thing->QueryFlag
		     (FLAGMODE => 'LINE',
		      INDEX => $Index,
		      FLAG => 'GNUP_MIN');
	    my $Max=$Thing->QueryFlag
		     (FLAGMODE => 'LINE',
		      INDEX => $Index,
		      FLAG => 'GNUP_MAX');
	    my $Smooth=$Thing->QueryFlag
		     (FLAGMODE => 'LINE',
		      INDEX => $Index,
		      FLAG => 'GNUP_SMOOTH');
	    my $Title=$Thing->QueryFlag
		     (FLAGMODE => 'LINE',
		      INDEX => $Index,
		      FLAG => 'GNUP_CURVETITLE');
	    $Title=MySubs::DoubleQuoteForShell($Title);
	    my $Pts=$Thing->QueryFlag
		     (FLAGMODE => 'LINE',
		      INDEX => $Index,
		      FLAG => 'GNUP_PTS');
	    my $Log=$Thing->QueryFlag
		     (FLAGMODE => 'LINE',
		      INDEX => $Index,
		      FLAG => 'GNUP_CURVELOG');
	    my $Listplot=$Thing->QueryFlag
		     (FLAGMODE => 'LINE',
		      INDEX => $Index,
		      FLAG => 'GNUP_LISTPLOT');
	    my $Command="$SubstVarName:funmake('SYMAXX_CreatePlotdata,['F=$SubstVarName";
	    if (defined($Indep) && $Indep ne '' && defined($Min) && $Min ne '' && defined($Max) && $Max ne ''){
		$Command.=",'RANGE=[$Indep,$Min,$Max]";};
	    if ($Pts){$Command.=",'PTS=$Pts";};
	    if ($Smooth){$Command.=",'SMOOTH=\"$Smooth\"";};
	    if ($Title){$Command.=",'CURVETITLE=\"$Title\"";};
	    if ($Log){$Command.=",'LINLOG=\"LOG\"";};
	    if ($Listplot){$Command.=",'LISTPLOT=TRUE";};
	    $Command.="])$NL";
	    push @Code,$Command;
	}; # if GNUPLOT_CreateData
	if ($Thing->QueryFlag
	    (FLAGMODE => 'LINE',
	     INDEX => $Index,
	     FLAG => 'GNUP_PLOTDATA')){		
	    my $Title=$Thing->QueryFlag
		(FLAGMODE => 'LINE',
		 INDEX => $Index,
		 FLAG => 'GNUP_DIAGTITLE');
	    $Title=MySubs::DoubleQuoteForShell($Title);
	    my $Scale=$Thing->QueryFlag
		(FLAGMODE => 'LINE',
		 INDEX => $Index,
		 FLAG => 'GNUP_DIAGSCALE');
	    my $LabelX=$Thing->QueryFlag
		(FLAGMODE => 'LINE',
		 INDEX => $Index,
		 FLAG => 'GNUP_DIAGLABELX');
	    my $LabelY=$Thing->QueryFlag
		(FLAGMODE => 'LINE',
		 INDEX => $Index,
		 FLAG => 'GNUP_DIAGLABELY');
	    my $LogX=$Thing->QueryFlag
		(FLAGMODE => 'LINE',
		 INDEX => $Index,
		 FLAG => 'GNUP_DIAGLOGX');
	    my $LogY=$Thing->QueryFlag
		(FLAGMODE => 'LINE',
		 INDEX => $Index,
		 FLAG => 'GNUP_DIAGLOGY');
	    my $GridX=$Thing->QueryFlag(FLAGMODE => 'LINE',INDEX => $Index,FLAG => 'GNUP_DIAGGRIDX');
	    my $GridY=$Thing->QueryFlag(FLAGMODE => 'LINE',INDEX => $Index,FLAG => 'GNUP_DIAGGRIDY');
	    my $RangeX=$Thing->QueryFlag(FLAGMODE => 'LINE',INDEX => $Index,FLAG => 'GNUP_DIAGRANGEX');
	    my $RangeY=$Thing->QueryFlag(FLAGMODE => 'LINE',INDEX => $Index,FLAG => 'GNUP_DIAGRANGEY');
	    my $Colour=$Thing->QueryFlag(FLAGMODE => 'LINE',INDEX => $Index,FLAG => 'GNUP_DIAGCOLOUR');
	    
	    my $Command="$SubstVarName:funmake('SYMAXX_GNUPLOT,['DATA=$SubstVarName";
	    if ($Title){$Command.=",'DIAGTITLE=\"$Title\"";};
	    if ($LogX){$Command.=",'LINLOGX=\"LOG\"";};
	    if ($LogY){$Command.=",'LINLOGY=\"LOG\"";};
	    if ($GridX){$Command.=",'GRIDX=1";};
	    if ($GridY){$Command.=",'GRIDY=1";};
	    if ($Colour){$Command.=",'COLOUR=\"color\"";};
	    if ($Scale){$Command.=",'SCALE=$Scale";};
	    if ($LabelX){$Command.=",'LABELX=\"$LabelX\"";};
	    if ($LabelY){$Command.=",'LABELY=\"$LabelY\"";};
	    if ($RangeX){$Command.=",'RANGEX=\"$RangeX\"";};
	    if ($RangeY){$Command.=",'RANGEY=\"$RangeY\"";};
	    $Command.="])$NL";
	    push @Code,$Command;
	}; # if GNUPLOT_PlotData
    };

    return @Code;
};
sub ApplyDispform{
    my %P=@_;
    my $ID=$P{ID} or die;
    my $Thing=$P{THING} or die;
    my $Index=$P{INDEX}; die unless defined $Index;
    my @Code=();    
    # Usually we convert to dispform, because it's easier to read for human beings than Maxima's internal notation.
    if (!$Thing->QueryFlag
	(FLAGMODE => 'LINE', 
	 INDEX => $Index, 
	 PATH => '',
	 FLAG => 'NODISPFORM')){
	push @Code,"block([SIMP:OFF],REF[$ID]:dispform(REF[$ID],ALL))$NL";
    };
    return @Code;
};	
sub StartNewEvaluation{
    my %P=@_;
    my $ID=$P{ID} or die;
    my $Thing=$P{THING} or die;
    my $Index=$P{INDEX}; die unless defined $Index;
    my @Code=();    
    # Create an initial command NIL
    # For the index 0, there will be only one command with Path ''.
    push @Code,"REF[$ID]:NIL$NL"; 
    push @Code,"SYMAXX_CLEAR_HISTORY()$NL"; 
    return @Code;
};
sub ControlLogic{
    my %P=@_;
    my $ID=$P{ID} or die;
    my $Thing=$P{THING} or die;
    my $Index=$P{INDEX}; die unless defined $Index;
    my $Path=$P{PATH}; die unless defined $Path;
    my $CommandCount=$P{COMMANDCOUNT}; die unless defined $CommandCount;
    my $EvalIt=0; my $RecallIt=0; my $SaveIt=0; my $StartNewEval=0;
    # The command has been changed, or it is a definition.
    if ($Thing->QueryEvaldirty || $Thing->QueryDefinition){
	$EvalIt=1;
		    };
    # If the RECALLNOEVAL-flag of this line is set,
    # we don't evaluate. We only recall it.
    # (A definition with RECALLNOEVAL set is simply ignored.)
    if ($Thing->QueryFlag
	(FLAGMODE=>'LINE',
	 INDEX => $Index,
	 FLAG => 'RECALLNOEVAL')){
	$EvalIt=0;
	if ($Path eq '' && !$Thing->QueryDefinition){$RecallIt=1;};
    };
    
    # The RECALLNOEVAL for the whole thing is even stronger:
    # (Recall will be reset later for the last result)
    if ($Thing->QueryFlag
	(FLAGMODE=>'THING',
	 FLAG => 'RECALLNOEVAL')){			    
	$RecallIt=0;$EvalIt=0;$SaveIt=0;
    };
			
    # The final result must be provided in the REF[]-array, otherwise it wouldn't be in the list.
    # Either by evaluation, or if that has not happened, by recalling.
    if (!$EvalIt && $Index==$CommandCount && $Path eq ''){
	$RecallIt=1;
    };			    
    
    # There is an evaluation and the command was dirty (for example not a definition).
    # Save the result.
    # 17.12.2001: Saveit may not be used, while the commands are still applied on the expression.
    # It involves dispform, which scrambles the order of the subexpressions!!
    if ($EvalIt && $Thing->QueryEvaldirty && $Path eq ''){$SaveIt=1;};
    
    # start a new evaluation also, if EvalIt is not set for the first line!
    if ($Index==0 && $Path eq ''){
	$StartNewEval=1;
    };
    return (EVALIT => $EvalIt,RECALLIT => $RecallIt,SAVEIT => $SaveIt,STARTNEWEVAL => $StartNewEval);
};
sub CheckLoglineForStartEval{
    my $LogLine=shift;
    return undef unless $LogLine;
#    print $LogLine."\n";

    if ($LogLine=~/
	$StartOfLine
	$MaybeBlanks
	\[SYMAXX-STARTEVAL\]   #Keyword
	$MaybeBlanks
	\-(\d+)\-          # -ID- of expression => $1
	$MaybeBlanks
	\-(\d+)\-          # -Index- => $2
	$MaybeBlanks
	\-(.*)\-           # -Path- => $3
	$MaybeBlanks
	$EndOfLine
	/x){
	return {ID => $1,INDEX => $2,PATH => $3};
    }; # if pattern found
    return undef;
};
sub CheckLoglineForEndEval{
    my $LogLine=shift;
    return undef unless $LogLine;

    if ($LogLine=~/
	$StartOfLine
	$MaybeBlanks
	\[SYMAXX-ENDEVAL\]   #Keyword
	\s*
	$EndOfLine
	/x){return 1;};
    return undef;
};
sub CheckLoglineForEvalSuccess{
    my $LogLine=shift;
    if ($LogLine=~/\[SYMAXX\-EVALSUCCESS\]/x){
	return 1;};
    return undef;
};
sub CheckLoglineForSymaxxError{		
    my $LogLine=shift;
    if ($LogLine=~/
	\[MAXIMA-ERROR-START\]
	$MaybeBlanks
	(.*)             #Error message => $1
	$MaybeBlanks
	\[MAXIMA-ERROR-END\]
	/x){
	if ($1!~/\[NO_ERROR\]/){
	    return &PostprocessErrorMessage($1);
	};
    }; 
    return undef;
};
sub PostprocessErrorMessage{
    my $Error=shift; return $Error unless $Error;
    my @ErrorArray=split(/\[MAXIMA-ERROR-ITEM\]/,$Error);
    $Error=shift(@ErrorArray); # the first entry is the error message.
    
    # Maxima uses different tokens : ~:M and ~%~M,...
    # Here, we don't distinguist between them, so convert all to the first one.
    $Error=~s/$Errortoken2/$Errortoken1_unquoted/g;
    $Error=~s/$Errortoken3/$Errortoken1_unquoted/g;
    $Error=~s/$Errortoken4/$Errortoken1_unquoted/g;
    $Error=~s/$Errortoken5/$Errortoken1_unquoted/g;

    while ($#ErrorArray >-1){
	my $Item=shift(@ErrorArray);
	$Error=~s/$Errortoken1/$Item/;
    };
    die unless $#ErrorArray==-1;
    $Error=~s/\s+/ /g;
    return $Error;
};
sub CheckLoglineForGeneralError{
    my $Error=shift; return 0 unless $Error;
    return $Error=~/.*an.*error.*quitting.*to.*debug.*this.*try.*debugmode/xi;
};
sub CheckLoglineForFatalError{
    my $Error=shift; return 0 unless $Error;
    return $Error=~/.*error.*caught.*fatal.*error.*memory.*may.*be.*damaged.*/xi;
};
